from ._mpe_utils.simple_env import SimpleEnv
from .scenarios.simple_speaker_listener import Scenario


class env(SimpleEnv):
    def __init__(self, seed=None, max_frames=100):
        scenario = Scenario()
        world = scenario.make_world()
        super(env, self).__init__(scenario, world, max_frames, seed)



import time
import random


def performance_benchmark(env):
    print("Starting performance benchmark")
    cycles = 0
    turn = 0
    _ = env.reset()
    start = time.time()
    end = 0

    while True:
        cycles += 1
        for agent in env.agent_order:  # step through every agent once with observe=True
            if 'legal_moves' in env.infos[agent]:
                action = random.choice(env.infos[agent]['legal_moves'])
            else:
                action = env.action_spaces[agent].sample()
            _ = env.step(action)
            turn += 1
            if all(env.dones.values()):
                _ = env.reset()
                break
        if time.time() - start > 5:
            end = time.time()
            break

    length = end - start

    turns_per_time = turn / length
    cycles_per_time = cycles / length
    print(str(turns_per_time) + " turns per second")
    print(str(cycles_per_time) + " cycles per second")
    print("Finished performance benchmark")



# from __future__ import absolute_import, division, print_function
from six.moves import range

import itertools
import copy
import numpy as np

from pettingzoo import AECEnv
from gym import spaces


class env(AECEnv):

    metadata = {'render.modes': ['human']}

    def __init__(self):
        super(env, self).__init__()

        self.ch = CheckersRules()
        self.num_agents = 2
        self.agents = ["player_{}".format(i) for i in range(self.num_agents)]
        self.agent_order = list(self.agents)

        self.action_spaces = {name: spaces.Discrete(8 * 8 * 73) for name in self.agents}
        self.observation_spaces = {name: spaces.Box(low=0, high=1, shape=(8, 8, 20), dtype=np.float32) for name in self.agents}
        self.observation = np.zeros((8, 8, 2))

        self.reset()

    def observe(self, agent):
        return self.observation[:, :, agent]

    def reset(self, observe=True):
        self.board = self.ch.initial_board()
        self.observation[:, :, 0] = np.array(self.ch.flat_board())
        self.observation[:, :, 1] = np.array(self.ch.flat_board())
        self.num_moves_max = 300
        self.num_moves = 0
        self.agent_selection = 0
        self.rewards = {name: 0 for name in self.agents}
        self.dones = {name: False for name in self.agents}
        self.infos = {name: {'legal_moves': []} for name in self.agents}
        self.infos["player_{}".format(self.agent_selection)]['legal_moves'] = self.ch.legal_moves()
        self.winner = -1

    def step(self, action, observe=True):
        self.num_moves += 1
        self.board, turn, last_moved_piece, moves, winner = self.ch.move(action[0], action[1])

        if turn == 'black':
            self.agent_selection = 0
        elif turn == 'white':
            self.agent_selection = 1
        else:
            raise ValueError

        self.observation[:, :, self.agent_selection] = np.array(self.ch.flat_board())

        print("After " + str(self.num_moves) + " moves: ")

        """
        self.ch.print_board()
        print(self.agent_selection)
        print(self.observe(self.agent_selection))
        print(self.observation)
        self.infos[self.agent_selection]['legal_moves']
        """

        self.infos["player_{}".format(self.agent_selection)]['legal_moves'] = moves

        if winner is None and self.num_moves > self.num_moves_max:
            print("Draw")
            self.winner = -1
            self.rewards[0] = 0
            self.rewards[1] = 0
        else:
            if winner == 'black':
                self.winner = 0
                self.rewards[0] = 1
                self.rewards[1] = -1
            elif winner == 'white':
                self.winner = 1
                self.rewards[0] = -1
                self.rewards[1] = 1
            else:
                pass

        if observe:
            next_observation = self.observe(self.agent_selection)
        else:
            next_observation = None
        return np.array(next_observation)

    def render(self, mode='human'):
        print(self.ch.flat_board())

    def close(self):
        pass


class CheckersRules:

    size = 8
    n_positions = int(size ** 2 // 2)
    n_per_row = int(size // 2)

    # TODO change players to top/bottom players
    all_players = ['black', 'white']
    all_piece_types = ['men', 'kings']

    # Converting to a flat representation of the board
    empty_square = 0
    black_man = 1
    black_king = 2
    white_man = 3
    white_king = 4

    # Directions
    pos2dir = ['sw', 'se', 'ne', 'nw']
    dir2del = [(+1, -1), (+1, +1), (-1, +1), (-1, -1)]

    # The directions a piece is allowed to move in
    legal_dirs = {
        'black': {
            'men': [0, 1],
            'kings': [0, 1, 2, 3],
        },
        'white': {
            'men': [2, 3],
            'kings': [0, 1, 2, 3],
        },
    }

    def __init__(self, board=None, turn='black', last_moved_piece=None, empty_corner=True):
        '''
        Args:
            empty_corner : bool
                If the upper left corner of the board should be used. Default to be False.
        '''
        # assert size == 8, 'Only supports size 8.'
        # assert turn in CheckersRules.all_players, 'It must be either `black` or `white`\'s turn'
        self.empty_corner = empty_corner

        # Game state
        self._board = board or self.initial_board()
        self._turn = turn
        self._last_moved_piece = None

        # LUT for the neighboring 4 squares in each directions respectively. None for a missing neighbor
        # XXX there is another way to find neighbors (consider [sq+4, sq+5, sq-4, sq-5])
        self.neighbors = {sq: [] for sq in range(self.n_positions)}
        for sq in range(self.n_positions):
            row, col = self.sq2pos(sq)
            # For each direction
            for di, (drow, dcol) in enumerate(CheckersRules.dir2del):
                next_row, next_col = row + drow, col + dcol
                # Out of bound
                if not (0 <= next_row < self.size and 0 <= next_col < self.size):
                    self.neighbors[sq].append(None)
                else:
                    self.neighbors[sq].append(self.pos2sq(next_row, next_col))

    @staticmethod
    def initial_board():
        '''Returns the initial configuration of the board'''
        # Black starts at the top of the board
        board = {
            'black': {
                'men': set(range(12)),
                'kings': set(),
            },
            'white': {
                'men': set(range(32 - 12, 32)),
                'kings': set(),
            },
        }
        return board

    @staticmethod
    def empty_board():
        board = {
            'black': {
                'men': set(),
                'kings': set(),
            },
            'white': {
                'men': set(),
                'kings': set(),
            },
        }
        return board

    @staticmethod
    def immutable_board(board):
        # TODO Bitboard representation?
        pieces = (
            frozenset(board['black']['men']),
            frozenset(board['black']['kings']),
            frozenset(board['white']['men']),
            frozenset(board['white']['kings']),
        )
        return pieces

    @staticmethod
    def board_equal(board1, board2):
        return CheckersRules.immutable_board(board1) == CheckersRules.immutable_board(board2)

    @property
    def board(self):
        return self._board

    @property
    def turn(self):
        return self._turn

    @property
    def last_moved_piece(self):
        return self._last_moved_piece

    def move(self, from_sq, to_sq, skip_check=False):
        '''Update the game state after the current player moves its piece from `from_sq` to `to_sq`. Reference: https://en.wikipedia.org/wiki/English_draughts#Rules
        Args:
            skip_check : bool
                If the move is chosen from results returned by `legal_moves()`, the legality check can be skipped for efficiency. Default to be False.
        '''
        if not skip_check:
            # Reject illegal moves
            assert (from_sq, to_sq) in self.legal_moves(), 'The move is not legal.'

        # The move is legal
        switch_turn = True
        # Move the piece
        for type in ['men', 'kings']:
            pieces = self._board[self._turn][type]
            if from_sq in pieces:
                pieces.remove(from_sq)
                pieces.add(to_sq)
                piece_type = type
                self._last_moved_piece = to_sq
                break
        else:
            assert False, 'A friendly piece must be moved.'

        # The move is a jump
        if to_sq not in self.neighbors[from_sq]:
            # Remove the captured piece
            to_row, to_col = self.sq2pos(to_sq)
            from_row, from_col = self.sq2pos(from_sq)
            capture_row, capture_col = (from_row + to_row) / 2, (from_col + to_col) / 2
            capture_sq = self.pos2sq(capture_row, capture_col)
            for type in ['men', 'kings']:
                pieces = self._board[self.adversary][type]
                if capture_sq in pieces:
                    pieces.remove(capture_sq)
                    break
            else:
                assert False, 'An opposing piece must be captured.'
            # Check for new available jumps for the moved piece before crowning a king
            jumps = self.available_jumps(self._turn, piece_type, to_sq)
            # Switch the turn, if there is no more jumps for the current player
            switch_turn = len(jumps) == 0

        # Crowning a king (must end the turn)
        if piece_type == 'men':
            # Kings row is at the bottom for black
            if self._turn == 'black' and self.n_positions - to_sq <= self.n_per_row:
                self._board[self._turn]['men'].remove(to_sq)
                self._board[self._turn]['kings'].add(to_sq)
            # Kings row is at the top for white
            if self._turn == 'white' and to_sq < self.n_per_row:
                self._board[self._turn]['men'].remove(to_sq)
                self._board[self._turn]['kings'].add(to_sq)

        if switch_turn:
            self._turn = self.adversary
            self._last_moved_piece = None

        # Check win/loss, winner is None before the game ends
        all_next_moves = self.legal_moves()
        if len(all_next_moves) == 0:
            winner = self.adversary
        else:
            winner = None
        return self.board, self.turn, self.last_moved_piece, all_next_moves, winner

    @property
    def adversary(self):
        return 'black' if self._turn == 'white' else 'white'

    def available_simple_moves(self, player, type, sq):
        simple_moves = []
        for di in CheckersRules.legal_dirs[player][type]:
            next_sq = self.neighbors[sq][di]
            # There is a neighboring square
            if next_sq is not None:
                # Check its occupancy
                if not self.check_occupancy(next_sq):
                    simple_moves.append(next_sq)
        return simple_moves

    def check_occupancy(self, sq, by_players=all_players):
        '''
        Return : bool
            True if `sq` is occupied.
        '''
        for player in by_players:
            for type in ['men', 'kings']:
                if sq in self._board[player][type]:
                    return True
        return False

    def available_jumps(self, player, type, sq):
        '''Returns the available jumps of `player`'s piece of `type` at `sq`.'''
        jumps = []
        adversary = 'black' if player == 'white' else 'white'
        for di in CheckersRules.legal_dirs[player][type]:
            capture_sq = self.neighbors[sq][di]
            # There is a neighboring square
            if capture_sq is not None:
                # The square is occupied by the adversary's piece
                if self.check_occupancy(capture_sq, [adversary]):
                    # Must jump over two squares in a single direction
                    next_sq = self.neighbors[capture_sq][di]
                    if next_sq is not None:
                        # The square is not occupied
                        if not self.check_occupancy(next_sq):
                            jumps.append(next_sq)
        return jumps

    def all_jumps(self):
        if self._last_moved_piece is None:
            jumps = []
            for type in ['men', 'kings']:
                for sq in self._board[self._turn][type]:
                    jumps += itertools.product([sq], self.available_jumps(self._turn, type, sq))
        else:
            piece_type = 'men' if self._last_moved_piece in self._board[self._turn]['men'] else 'kings'
            jumps = itertools.product([self._last_moved_piece], self.available_jumps(self._turn, piece_type, self._last_moved_piece))
        return list(jumps)

    def legal_moves(self):
        '''Returns all legal moves of the current `player`.'''
        all_moves = self.all_jumps()
        # Jumps are mandatory
        if 0 < len(all_moves):
            return all_moves
        # No jumps available
        for type in ['men', 'kings']:
            for sq in self._board[self._turn][type]:
                all_moves += itertools.product([sq], self.available_simple_moves(self._turn, type, sq))
        return all_moves

    def pos2sq(self, row, col):
        if self.empty_corner and row % 2 == 0:
            # Even rows starts with an empty square
            col -= 1
        elif not self.empty_corner and row % 2 == 1:
            # Odd rows starts with an empty square
            col -= 1
        col /= 2
        return int(row * (self.size / 2) + col)

    def sq2pos(self, sq):
        row, col = int(sq // (self.size / 2)), int(sq % (self.size / 2))
        col *= 2
        if self.empty_corner and row % 2 == 0:
            # Even rows starts with an empty square
            col += 1
        elif not self.empty_corner and row % 2 == 1:
            # Odd rows starts with an empty square
            col += 1
        return row, col

    def flat_board(self):
        # Empty board
        board = np.ones((self.size, self.size), dtype='int') * CheckersRules.empty_square
        # Place the pieces
        for sq in self._board['black']['men']:
            row, col = self.sq2pos(sq)
            board[row][col] = CheckersRules.black_man
        for sq in self._board['black']['kings']:
            row, col = self.sq2pos(sq)
            board[row][col] = CheckersRules.black_king
        for sq in self._board['white']['men']:
            row, col = self.sq2pos(sq)
            board[row][col] = CheckersRules.white_man
        for sq in self._board['white']['kings']:
            row, col = self.sq2pos(sq)
            board[row][col] = CheckersRules.white_king
        return board

    def print_board(self):
        # Symbols
        # print(self.flat_board())
        empty_square = '_'
        empty_playable_square = '.'
        black_man = 'b'
        black_king = 'B'
        white_man = 'w'
        white_king = 'W'
        symbols = [empty_playable_square, black_man, black_king, white_man, white_king]
        # Print board
        for i, row in enumerate(self.flat_board()):
            for j, col in enumerate(row):
                if ((i + self.empty_corner) % 2 + j) % 2 == 1:
                    # Not playable squares
                    print(empty_square, end='')
                else:
                    print(symbols[col], end='')
            print()

    def print_empty_board(self):
        '''Display the standard representation of the board with squares:
        __00__01__02__03
        04__05__06__07__
        __08__09__10__11
        12__13__14__15__
        __16__17__18__19
        20__21__22__23__
        __24__25__26__27
        28__29__30__31__
        '''
        board = -1 * np.ones((self.size, self.size), dtype='int')
        # Print board
        for sq in range(self.n_positions):
            board[self.sq2pos(sq)] = sq
        for row in board:
            for col in row:
                print('__' if col < 0 else '%02i' % col, end='')
            print()

    def save_state(self):
        return copy.deepcopy(self.board), self.turn, self.last_moved_piece

    def restore_state(self, state):
        board, turn, last_moved_piece = state
        self._board = copy.deepcopy(board)
        self._turn = turn
        self._last_moved_piece = last_moved_piece


"""
# Human keyboard player
def keyboard_player_move(board, last_moved_piece):
    '''A player that uses keyboard to select moves.'''
    if last_moved_piece is None:
        input_str = input('* move `from_square, to_square`: ')
    else:
        input_str = input('* move `%i, to_square`: ' % last_moved_piece)
    from_sq, to_sq = map(int, input_str.strip().split(','))
    return from_sq, to_sq

"""



from ._mpe_utils.simple_env import SimpleEnv
from .scenarios.simple_world_comm import Scenario


class env(SimpleEnv):
    def __init__(self, seed=None, num_good=2, num_adversaries=4, num_obstacles=1, num_food=2, num_forests=2, max_frames=100):
        scenario = Scenario()
        world = scenario.make_world(num_good, num_adversaries, num_obstacles, num_food, num_forests)
        super(env, self).__init__(scenario, world, max_frames, seed)



from . import simple_adversary as simple_adversary_v0
from . import simple_crypto as simple_crypto_v0
from . import simple_push as simple_push_v0
from . import simple_reference as simple_reference_v0
from . import simple_speaker_listener as simple_speaker_listener_v0
from . import simple_spread as simple_spread_v0
from . import simple_tag as simple_tag_v0
from . import simple_world_comm as simple_world_comm_v0
from . import simple as simple_v0



import copy
import math
import numpy as np
from gym import spaces
from gym.utils import seeding
import Box2D
from Box2D.b2 import (circleShape, contactListener, edgeShape, fixtureDef, polygonShape,
                      revoluteJointDef)
from .. import Agent
from pettingzoo.utils import EnvLogger

MAX_AGENTS = 40

FPS = 50
SCALE = 30.0  # affects how fast-paced the game is, forces should be adjusted as well

MOTORS_TORQUE = 80
SPEED_HIP = 4
SPEED_KNEE = 6
LIDAR_RANGE = 160 / SCALE

INITIAL_RANDOM = 5

HULL_POLY = [(-30, +9), (+6, +9), (+34, +1), (+34, -8), (-30, -8)]
LEG_DOWN = -8 / SCALE
LEG_W, LEG_H = 8 / SCALE, 34 / SCALE

PACKAGE_POLY = [(-120, 5), (120, 5), (120, -5), (-120, -5)]

PACKAGE_LENGTH = 240

VIEWPORT_W = 600
VIEWPORT_H = 400

TERRAIN_STEP = 14 / SCALE
TERRAIN_LENGTH = 200  # in steps
TERRAIN_HEIGHT = VIEWPORT_H / SCALE / 4
TERRAIN_GRASS = 10  # low long are grass spots, in steps
TERRAIN_STARTPAD = 20  # in steps
FRICTION = 2.5

WALKER_SEPERATION = 10  # in steps


class ContactDetector(contactListener):

    def __init__(self, env):
        contactListener.__init__(self)
        self.env = env

    def BeginContact(self, contact):
        # if walkers fall on ground
        for i, walker in enumerate(self.env.walkers):
            if walker.hull == contact.fixtureA.body:
                if self.env.package != contact.fixtureB.body:
                    self.env.fallen_walkers[i] = True
            if walker.hull == contact.fixtureB.body:
                if self.env.package != contact.fixtureA.body:
                    self.env.fallen_walkers[i] = True

        # if package is on the ground
        if self.env.package == contact.fixtureA.body:
            if contact.fixtureB.body not in [w.hull for w in self.env.walkers]:
                self.env.game_over = True
        if self.env.package == contact.fixtureB.body:
            if contact.fixtureA.body not in [w.hull for w in self.env.walkers]:
                self.env.game_over = True

            #    self.env.game_over = True
        for walker in self.env.walkers:
            for leg in [walker.legs[1], walker.legs[3]]:
                if leg in [contact.fixtureA.body, contact.fixtureB.body]:
                    leg.ground_contact = True

    def EndContact(self, contact):
        for walker in self.env.walkers:
            for leg in [walker.legs[1], walker.legs[3]]:
                if leg in [contact.fixtureA.body, contact.fixtureB.body]:
                    leg.ground_contact = False


class BipedalWalker(Agent):

    def __init__(self, world, init_x=TERRAIN_STEP * TERRAIN_STARTPAD / 2,
                 init_y=TERRAIN_HEIGHT + 2 * LEG_H, n_walkers=2, seed=None):
        self.world = world
        self._n_walkers = n_walkers
        self.hull = None
        self.init_x = init_x
        self.init_y = init_y
        self._seed(seed)

    def _destroy(self):
        if not self.hull:
            return
        self.world.DestroyBody(self.hull)
        self.hull = None
        for leg in self.legs:
            self.world.DestroyBody(leg)
        self.legs = []
        self.joints = []

    def _seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def _reset(self):
        self._destroy()

        init_x = self.init_x
        init_y = self.init_y
        self.hull = self.world.CreateDynamicBody(
            position=(init_x, init_y),
            fixtures=fixtureDef(
                shape=polygonShape(
                    vertices=[(x / SCALE, y / SCALE) for x, y in HULL_POLY]),
                density=5.0,
                friction=0.1,
                categoryBits=0x002,
                # maskBits=(0x001 & 0x002),  # collide only with ground
                restitution=0.0)  # 0.99 bouncy
        )
        self.hull.color1 = (0.5, 0.4, 0.9)
        self.hull.color2 = (0.3, 0.3, 0.5)
        self.hull.ApplyForceToCenter((self.np_random.uniform(-INITIAL_RANDOM, INITIAL_RANDOM), 0),
                                     True)

        self.legs = []
        self.joints = []
        for i in [-1, +1]:
            leg = self.world.CreateDynamicBody(
                position=(init_x, init_y - LEG_H / 2 - LEG_DOWN),
                angle=(i * 0.05),
                fixtures=fixtureDef(shape=polygonShape(box=(LEG_W / 2, LEG_H / 2)), density=1.0,
                                    restitution=0.0, categoryBits=0x002,
                                    maskBits=0x001)  # collide with ground only
            )
            leg.color1 = (0.6 - i / 10., 0.3 - i / 10., 0.5 - i / 10.)
            leg.color2 = (0.4 - i / 10., 0.2 - i / 10., 0.3 - i / 10.)
            rjd = revoluteJointDef(
                bodyA=self.hull,
                bodyB=leg,
                localAnchorA=(0, LEG_DOWN),
                localAnchorB=(0, LEG_H / 2),
                enableMotor=True,
                enableLimit=True,
                maxMotorTorque=MOTORS_TORQUE,
                motorSpeed=i,
                lowerAngle=-0.8,
                upperAngle=1.1,)
            self.legs.append(leg)
            self.joints.append(self.world.CreateJoint(rjd))

            lower = self.world.CreateDynamicBody(
                position=(init_x, init_y - LEG_H * 3 / 2 - LEG_DOWN), angle=(i * 0.05),
                fixtures=fixtureDef(shape=polygonShape(box=(0.8 * LEG_W / 2, LEG_H / 2)),
                                    density=1.0, restitution=0.0, categoryBits=0x0020,
                                    maskBits=0x001))
            lower.color1 = (0.6 - i / 10., 0.3 - i / 10., 0.5 - i / 10.)
            lower.color2 = (0.4 - i / 10., 0.2 - i / 10., 0.3 - i / 10.)
            rjd = revoluteJointDef(
                bodyA=leg,
                bodyB=lower,
                localAnchorA=(0, -LEG_H / 2),
                localAnchorB=(0, LEG_H / 2),
                enableMotor=True,
                enableLimit=True,
                maxMotorTorque=MOTORS_TORQUE,
                motorSpeed=1,
                lowerAngle=-1.6,
                upperAngle=-0.1,)
            lower.ground_contact = False
            self.legs.append(lower)
            self.joints.append(self.world.CreateJoint(rjd))

        self.drawlist = self.legs + [self.hull]

        class LidarCallback(Box2D.b2.rayCastCallback):

            def ReportFixture(self, fixture, point, normal, fraction):
                if (fixture.filterData.categoryBits & 1) == 0:
                    return 1
                self.p2 = point
                self.fraction = fraction
                return 0

        self.lidar = [LidarCallback() for _ in range(10)]

    def apply_action(self, action):

        self.joints[0].motorSpeed = float(SPEED_HIP * np.sign(action[0]))
        self.joints[0].maxMotorTorque = float(
            MOTORS_TORQUE * np.clip(np.abs(action[0]), 0, 1))
        self.joints[1].motorSpeed = float(SPEED_KNEE * np.sign(action[1]))
        self.joints[1].maxMotorTorque = float(
            MOTORS_TORQUE * np.clip(np.abs(action[1]), 0, 1))
        self.joints[2].motorSpeed = float(SPEED_HIP * np.sign(action[2]))
        self.joints[2].maxMotorTorque = float(
            MOTORS_TORQUE * np.clip(np.abs(action[2]), 0, 1))
        self.joints[3].motorSpeed = float(SPEED_KNEE * np.sign(action[3]))
        self.joints[3].maxMotorTorque = float(
            MOTORS_TORQUE * np.clip(np.abs(action[3]), 0, 1))

    def get_observation(self):
        pos = self.hull.position
        vel = self.hull.linearVelocity

        for i in range(10):
            self.lidar[i].fraction = 1.0
            self.lidar[i].p1 = pos
            self.lidar[i].p2 = (pos[0] + math.sin(1.5 * i / 10.0) * LIDAR_RANGE,
                                pos[1] - math.cos(1.5 * i / 10.0) * LIDAR_RANGE)
            self.world.RayCast(
                self.lidar[i], self.lidar[i].p1, self.lidar[i].p2)

        state = [
            # Normal angles up to 0.5 here, but sure more is possible.
            self.hull.angle,
            2.0 * self.hull.angularVelocity / FPS,
            # Normalized to get -1..1 range
            0.3 * vel.x * (VIEWPORT_W / SCALE) / FPS,
            0.3 * vel.y * (VIEWPORT_H / SCALE) / FPS,
            self.joints[0].
            # This will give 1.1 on high up, but it's still OK (and there should be spikes on hiting the ground, that's normal too)
            angle,
            self.joints[0].speed / SPEED_HIP,
            self.joints[1].angle + 1.0,
            self.joints[1].speed / SPEED_KNEE,
            1.0 if self.legs[1].ground_contact else 0.0,
            self.joints[2].angle,
            self.joints[2].speed / SPEED_HIP,
            self.joints[3].angle + 1.0,
            self.joints[3].speed / SPEED_KNEE,
            1.0 if self.legs[3].ground_contact else 0.0
        ]

        state += [l.fraction for l in self.lidar]
        assert len(state) == 24

        return state

    @property
    def observation_space(self):
        # 24 original obs (joints, etc), 2 displacement obs for each neighboring walker, 3 for package
        return spaces.Box(low=np.float32(-LIDAR_RANGE), high=np.float32(LIDAR_RANGE), shape=(24 + 4 + 3,), dtype=np.float32)

    @property
    def action_space(self):
        return spaces.Box(low=np.float32(-1), high=np.float32(1), shape=(4,), dtype=np.float32)


class MultiWalkerEnv():

    metadata = {'render.modes': [
        'human', 'rgb_array'], 'video.frames_per_second': FPS}

    hardcore = False

    def __init__(self, seed=None, n_walkers=3, position_noise=1e-3, angle_noise=1e-3, reward_mech='local',
                 forward_reward=1.0, fall_reward=-100.0, drop_reward=-100.0, terminate_on_fall=True, max_frames=500):
        # reward_mech is 'global' for cooperative game (same reward for every agent)
        """
            n_walkers: number of bipedal walkers in environment
            position_noise: noise applied to agent positional sensor observations
            angle_noise: noise applied to agent rotational sensor observations
            reward_mech: whether all agents are rewarded equal amounts or singular agent is rewarded
            forward_reward: reward applied for an agent standing, scaled by agent's x coordinate
            fall_reward: reward applied when an agent falls down
            drop_reward: reward applied for each fallen walker in environment
            terminate_on_fall: toggles whether agent is done if it falls down
            max_frames: after max_frames steps all agents will return done
        """

        self.n_walkers = n_walkers
        self.position_noise = position_noise
        self.angle_noise = angle_noise
        self._reward_mech = reward_mech
        self.forward_reward = forward_reward
        self.fall_reward = fall_reward
        self.drop_reward = drop_reward
        self.terminate_on_fall = terminate_on_fall
        self.seed_val = seed
        self.seed(seed=seed)
        self.setup()
        self.agent_list = list(range(self.n_walkers))
        self.last_rewards = [0 for _ in range(self.n_walkers)]
        self.last_dones = [False for _ in range(self.n_walkers)]
        self.last_obs = [None for _ in range(self.n_walkers)]
        self.max_frames = max_frames
        self.frames = 0

    def get_param_values(self):
        return self.__dict__

    def setup(self):
        self.viewer = None

        self.world = Box2D.b2World()
        self.terrain = None

        init_x = TERRAIN_STEP * TERRAIN_STARTPAD / 2
        init_y = TERRAIN_HEIGHT + 2 * LEG_H
        self.start_x = [
            init_x + WALKER_SEPERATION * i * TERRAIN_STEP for i in range(self.n_walkers)
        ]
        self.walkers = [
            BipedalWalker(self.world, init_x=sx,
                          init_y=init_y, seed=self.seed_val)
            for sx in self.start_x
        ]
        self.num_agents = len(self.walkers)
        self.observation_space = [
            agent.observation_space for agent in self.walkers]
        self.action_space = [agent.action_space for agent in self.walkers]

        self.package_scale = self.n_walkers / 1.75
        self.package_length = PACKAGE_LENGTH / SCALE * self.package_scale

        self.total_agents = self.n_walkers

        self.prev_shaping = np.zeros(self.n_walkers)
        self.prev_package_shaping = 0.0

        self.terrain_length = int(TERRAIN_LENGTH * self.n_walkers * 1 / 8.)

        self.reset()

    @property
    def agents(self):
        return self.walkers

    @property
    def reward_mech(self):
        return self._reward_mech

    def seed(self, seed=None):
        self.np_random, seed_ = seeding.np_random(seed)
        return [seed_]

    def _destroy(self):
        if not self.terrain:
            return
        self.world.contactListener = None
        for t in self.terrain:
            self.world.DestroyBody(t)
        self.terrain = []
        self.world.DestroyBody(self.package)
        self.package = None

        for walker in self.walkers:
            walker._destroy()

    def close(self):
        if self.viewer is not None:
            self.viewer.close()
            self.viewer = None
        else:
            EnvLogger.warn_close_unrendered_env()

    def reset(self):
        self._destroy()
        self.world.contactListener_bug_workaround = ContactDetector(self)
        self.world.contactListener = self.world.contactListener_bug_workaround
        self.game_over = False
        self.fallen_walkers = np.zeros(self.n_walkers, dtype=np.bool)
        self.prev_shaping = np.zeros(self.n_walkers)
        self.prev_package_shaping = 0.0
        self.scroll = 0.0
        self.lidar_render = 0

        self._generate_package()
        self._generate_terrain(self.hardcore)
        self._generate_clouds()

        self.drawlist = copy.copy(self.terrain)

        self.drawlist += [self.package]

        for walker in self.walkers:
            walker._reset()
            self.drawlist += walker.legs
            self.drawlist += [walker.hull]
        r, d, o = self.scroll_subroutine()
        self.last_rewards = [0 for _ in range(self.n_walkers)]
        self.last_dones = [False for _ in range(self.n_walkers)]
        self.last_obs = o
        self.frames = 0

        return self.observe(0)

    def scroll_subroutine(self):
        xpos = np.zeros(self.n_walkers)
        obs = []
        done = False
        rewards = np.zeros(self.n_walkers)

        for i in range(self.n_walkers):
            pos = self.walkers[i].hull.position
            x, y = pos.x, pos.y
            xpos[i] = x

            wobs = self.walkers[i].get_observation()
            nobs = []
            for j in [i - 1, i + 1]:
                # if no neighbor (for edge walkers)
                if j < 0 or j == self.n_walkers:
                    nobs.append(0.0)
                    nobs.append(0.0)
                else:
                    xm = (self.walkers[j].hull.position.x - x) / self.package_length
                    ym = (self.walkers[j].hull.position.y - y) / self.package_length
                    nobs.append(self.np_random.normal(xm, self.position_noise))
                    nobs.append(self.np_random.normal(ym, self.position_noise))
            xd = (self.package.position.x - x) / self.package_length
            yd = (self.package.position.y - y) / self.package_length
            nobs.append(self.np_random.normal(xd, self.position_noise))
            nobs.append(self.np_random.normal(yd, self.position_noise))
            nobs.append(self.np_random.normal(self.package.angle, self.angle_noise))
            obs.append(np.array(wobs + nobs))

            # shaping = 130 * pos[0] / SCALE
            shaping = 0.0
            shaping -= 5.0 * abs(wobs[0])
            rewards[i] = shaping - self.prev_shaping[i]
            self.prev_shaping[i] = shaping

        package_shaping = self.forward_reward * 130 * self.package.position.x / SCALE
        rewards += (package_shaping - self.prev_package_shaping)
        self.prev_package_shaping = package_shaping

        self.scroll = xpos.mean() - VIEWPORT_W / SCALE / 5 - (self.n_walkers - 1) * \
            WALKER_SEPERATION * TERRAIN_STEP

        done = False
        if self.game_over or pos[0] < 0:
            rewards += self.drop_reward
            done = True
        if pos[0] > (self.terrain_length - TERRAIN_GRASS) * TERRAIN_STEP:
            done = True
        rewards += self.fall_reward * self.fallen_walkers
        if self.terminate_on_fall and np.sum(self.fallen_walkers) > 0:
            done = True

        return rewards, done, obs

    def step(self, action, agent_id, is_last):
        # action is array of size 4
        action = action.reshape(4)
        self.walkers[agent_id].apply_action(action)
        self.world.Step(1.0 / FPS, 6 * 30, 2 * 30)
        self.frames = self.frames + 1
        if is_last:
            rewards, done, mod_obs = self.scroll_subroutine()
            self.last_obs[agent_id] = mod_obs[agent_id]
            if self.reward_mech == 'local':
                self.last_rewards = rewards
            else:
                self.last_rewards = [rewards.mean()
                                     for _ in range(self.n_walkers)]
            self.last_dones = [done for _ in range(self.n_walkers)]

    def get_last_rewards(self):
        return dict(zip(list(range(self.n_walkers)), map(lambda r: np.float64(r), self.last_rewards)))

    def get_last_dones(self):
        return dict(zip(self.agent_list, self.last_dones))

    def get_last_obs(self):
        return dict(zip(list(range(self.n_walkers)), [walker.get_observation() for walker in self.walkers]))

    def observe(self, agent):
        o = self.last_obs[agent]
        o = np.array(o)
        return o

    def render(self, mode='human', close=False):
        if close:
            if self.viewer is not None:
                self.viewer.close()
                self.viewer = None
            return

        render_scale = 0.75

        from gym.envs.classic_control import rendering
        if self.viewer is None:
            self.viewer = rendering.Viewer(VIEWPORT_W, VIEWPORT_H)
        self.viewer.set_bounds(self.scroll,
                               VIEWPORT_W / SCALE * self.package_scale * render_scale + self.scroll,
                               0, VIEWPORT_H / SCALE * self.package_scale * render_scale)

        self.viewer.draw_polygon([
            (self.scroll, 0),
            (self.scroll + VIEWPORT_W * self.package_scale / SCALE * render_scale, 0),
            (self.scroll + VIEWPORT_W * self.package_scale / SCALE * render_scale,
             VIEWPORT_H / SCALE * self.package_scale * render_scale),
            (self.scroll, VIEWPORT_H / SCALE * self.package_scale * render_scale),
        ], color=(0.9, 0.9, 1.0))
        for poly, x1, x2 in self.cloud_poly:
            if x2 < self.scroll / 2:
                continue
            if x1 > self.scroll / 2 + VIEWPORT_W / SCALE * self.package_scale:
                continue
            self.viewer.draw_polygon(
                [(p[0] + self.scroll / 2, p[1]) for p in poly], color=(1, 1, 1))
        for poly, color in self.terrain_poly:
            if poly[1][0] < self.scroll:
                continue
            if poly[0][0] > self.scroll + VIEWPORT_W / SCALE * self.package_scale:
                continue
            self.viewer.draw_polygon(poly, color=color)

        self.lidar_render = (self.lidar_render + 1) % 100
        i = self.lidar_render
        for walker in self.walkers:
            if i < 2 * len(walker.lidar):
                l = walker.lidar[i] if i < len(walker.lidar) else walker.lidar[len(walker.lidar) - i - 1]
                self.viewer.draw_polyline(
                    [l.p1, l.p2], color=(1, 0, 0), linewidth=1)

        for obj in self.drawlist:
            for f in obj.fixtures:
                trans = f.body.transform
                if type(f.shape) is circleShape:
                    t = rendering.Transform(translation=trans * f.shape.pos)
                    self.viewer.draw_circle(
                        f.shape.radius, 30, color=obj.color1).add_attr(t)
                    self.viewer.draw_circle(f.shape.radius, 30, color=obj.color2, filled=False,
                                            linewidth=2).add_attr(t)
                else:
                    path = [trans * v for v in f.shape.vertices]
                    self.viewer.draw_polygon(path, color=obj.color1)
                    path.append(path[0])
                    self.viewer.draw_polyline(
                        path, color=obj.color2, linewidth=2)

        flagy1 = TERRAIN_HEIGHT
        flagy2 = flagy1 + 50 / SCALE
        x = TERRAIN_STEP * 3
        self.viewer.draw_polyline(
            [(x, flagy1), (x, flagy2)], color=(0, 0, 0), linewidth=2)
        f = [(x, flagy2), (x, flagy2 - 10 / SCALE),
             (x + 25 / SCALE, flagy2 - 5 / SCALE)]
        self.viewer.draw_polygon(f, color=(0.9, 0.2, 0))
        self.viewer.draw_polyline(f + [f[0]], color=(0, 0, 0), linewidth=2)

        return self.viewer.render(return_rgb_array=mode == 'rgb_array')

    def _generate_package(self):
        init_x = np.mean(self.start_x)
        init_y = TERRAIN_HEIGHT + 3 * LEG_H
        self.package = self.world.CreateDynamicBody(
            position=(init_x, init_y),
            fixtures=fixtureDef(
                shape=polygonShape(vertices=[(x * self.package_scale / SCALE, y / SCALE)
                                             for x, y in PACKAGE_POLY]),
                density=1.0,
                friction=0.5,
                categoryBits=0x004,
                # maskBits=0x001,  # collide only with ground
                restitution=0.0)  # 0.99 bouncy
        )
        self.package.color1 = (0.5, 0.4, 0.9)
        self.package.color2 = (0.3, 0.3, 0.5)

    def _generate_terrain(self, hardcore):
        GRASS, STUMP, STAIRS, PIT, _STATES_ = range(5)
        state = GRASS
        velocity = 0.0
        y = TERRAIN_HEIGHT
        counter = TERRAIN_STARTPAD
        oneshot = False
        self.terrain = []
        self.terrain_x = []
        self.terrain_y = []
        for i in range(self.terrain_length):
            x = i * TERRAIN_STEP
            self.terrain_x.append(x)

            if state == GRASS and not oneshot:
                velocity = 0.8 * velocity + 0.01 * np.sign(TERRAIN_HEIGHT - y)
                if i > TERRAIN_STARTPAD:
                    velocity += self.np_random.uniform(-1, 1) / SCALE
                y += velocity

            elif state == PIT and oneshot:
                counter = self.np_random.randint(3, 5)
                poly = [
                    (x, y),
                    (x + TERRAIN_STEP, y),
                    (x + TERRAIN_STEP, y - 4 * TERRAIN_STEP),
                    (x, y - 4 * TERRAIN_STEP),
                ]
                t = self.world.CreateStaticBody(fixtures=fixtureDef(
                    shape=polygonShape(vertices=poly), friction=FRICTION))
                t.color1, t.color2 = (1, 1, 1), (0.6, 0.6, 0.6)
                self.terrain.append(t)
                t = self.world.CreateStaticBody(fixtures=fixtureDef(shape=polygonShape(
                    vertices=[(p[0] + TERRAIN_STEP * counter, p[1]) for p in poly]),
                    friction=FRICTION))
                t.color1, t.color2 = (1, 1, 1), (0.6, 0.6, 0.6)
                self.terrain.append(t)
                counter += 2
                original_y = y

            elif state == PIT and not oneshot:
                y = original_y
                if counter > 1:
                    y -= 4 * TERRAIN_STEP

            elif state == STUMP and oneshot:
                counter = self.np_random.randint(1, 3)
                poly = [
                    (x, y),
                    (x + counter * TERRAIN_STEP, y),
                    (x + counter * TERRAIN_STEP, y + counter * TERRAIN_STEP),
                    (x, y + counter * TERRAIN_STEP),
                ]
                t = self.world.CreateStaticBody(fixtures=fixtureDef(
                    shape=polygonShape(vertices=poly), friction=FRICTION))
                t.color1, t.color2 = (1, 1, 1), (0.6, 0.6, 0.6)
                self.terrain.append(t)

            elif state == STAIRS and oneshot:
                stair_height = +1 if self.np_random.rand() > 0.5 else -1
                stair_width = self.np_random.randint(4, 5)
                stair_steps = self.np_random.randint(3, 5)
                original_y = y
                for s in range(stair_steps):
                    poly = [
                        (x + (s * stair_width) * TERRAIN_STEP,
                         y + (s * stair_height) * TERRAIN_STEP),
                        (x + ((1 + s) * stair_width) * TERRAIN_STEP,
                         y + (s * stair_height) * TERRAIN_STEP),
                        (x + ((1 + s) * stair_width) * TERRAIN_STEP,
                         y + (-1 + s * stair_height) * TERRAIN_STEP),
                        (x + (s * stair_width) * TERRAIN_STEP,
                         y + (-1 + s * stair_height) * TERRAIN_STEP),
                    ]
                    t = self.world.CreateStaticBody(fixtures=fixtureDef(
                        shape=polygonShape(vertices=poly), friction=FRICTION))
                    t.color1, t.color2 = (1, 1, 1), (0.6, 0.6, 0.6)
                    self.terrain.append(t)
                counter = stair_steps * stair_width

            elif state == STAIRS and not oneshot:
                s = stair_steps * stair_width - counter - stair_height
                n = s / stair_width
                y = original_y + (n * stair_height) * TERRAIN_STEP

            oneshot = False
            self.terrain_y.append(y)
            counter -= 1
            if counter == 0:
                counter = self.np_random.randint(
                    TERRAIN_GRASS / 2, TERRAIN_GRASS)
                if state == GRASS and hardcore:
                    state = self.np_random.randint(1, _STATES_)
                    oneshot = True
                else:
                    state = GRASS
                    oneshot = True

        self.terrain_poly = []
        for i in range(self.terrain_length - 1):
            poly = [(self.terrain_x[i], self.terrain_y[i]),
                    (self.terrain_x[i + 1], self.terrain_y[i + 1])]
            t = self.world.CreateStaticBody(fixtures=fixtureDef(
                shape=edgeShape(vertices=poly),
                friction=FRICTION,
                categoryBits=0x0001,))
            color = (0.3, 1.0 if i % 2 == 0 else 0.8, 0.3)
            t.color1 = color
            t.color2 = color
            self.terrain.append(t)
            color = (0.4, 0.6, 0.3)
            poly += [(poly[1][0], 0), (poly[0][0], 0)]
            self.terrain_poly.append((poly, color))
        self.terrain.reverse()

    def _generate_clouds(self):
        # Sorry for the clouds, couldn't resist
        self.cloud_poly = []
        for i in range(self.terrain_length // 20):
            x = self.np_random.uniform(0, self.terrain_length) * TERRAIN_STEP
            y = VIEWPORT_H / SCALE * 3 / 4
            poly = [(x + 15 * TERRAIN_STEP * math.sin(3.14 * 2 * a / 5) + self.np_random.uniform(
                0, 5 * TERRAIN_STEP), y + 5 * TERRAIN_STEP * math.cos(3.14 * 2 * a / 5) + self.np_random.uniform(0, 5 * TERRAIN_STEP)) for a in range(5)]
            x1 = min([p[0] for p in poly])
            x2 = max([p[0] for p in poly])
            self.cloud_poly.append((poly, x1, x2))



import pygame
import pygame as pg
import os
import random
import gym
import numpy as np

# Define some colors

PLAYER_SPEED = 30.0
PLAYER_ROT_SPEED = 20.0
PLAYER_IMG = "manBlue_gun.png"
PLAYER_HIT_RECT = pg.Rect(0, 0, 35, 35)
TILESIZE = 2

BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
GREEN = (0, 255, 0)
RED = (255, 0, 0)

_image_library = {}
vec = pygame.math.Vector2


def get_image(path):
    global _image_library
    image = _image_library.get(path)
    if image is None:
        canonicalized_path = path.replace("/", os.sep).replace("\\", os.sep)
        image = pygame.image.load(canonicalized_path)
        _image_library[path] = image
    return image


def get_small_random_value():
    # generates a small random value between [0, 1/100)
    return (1 / 100) * np.random.rand()


class Block(pygame.sprite.Sprite):
    """
    This class represents the ball.
    It derives from the "Sprite" class in Pygame.
    """

    def __init__(self, color=(250, 250, 0), width=20, height=20):
        """ Constructor. Pass in the color of the block,
        and its size. """

        # Call the parent class (Sprite) constructor
        super().__init__()

        # Create an image of the block, and fill it with a color.
        # This could also be an image loaded from the disk.
        self.image = pygame.Surface([width, height])
        self.image.fill(color)
        # Fetch the rectangle object that has the dimensions of the image
        # image.
        # Update the position of this object by setting the values
        # of rect.x and rect.y
        self.rect = self.image.get_rect()
        self.dim = self.rect.size

    def reset_pos(self):
        self.rect.y = random.randrange(0, 20)
        self.rect.x = random.randrange(0, self.screen_width[0])
        # print(self.rect.y, self.rect.x)

    def update(self, pos, corner):
        """
        rect corner numbers
        1-----2
        |     |
        4-----3
        """
        self.rect.bottomright = (pos.x, pos.y)


class agent1(pygame.sprite.Sprite):
    """
    This class represents the ball
    """

    def __init__(self, _screen_width, x, y, speed=20):
        """ Constructor. Pass in the color of the block,
        and its x and y position. """
        # Call the parent class (Sprite) constructor
        super().__init__()

        self.image = get_image("agent1.jpg")
        self.base_image = get_image("agent1.jpg")
        self.screen_width = _screen_width
        self.rect = self.image.get_rect()
        self.dim = self.rect.size

        self.speed_val = speed
        self.vel = vec(
            int(self.speed_val * np.cos(np.pi / 4)),
            int(self.speed_val * np.sin(np.pi / 4)),
        )
        self.rot = 0
        self.bounce_randomness = 1
        self.pos = vec(x, y) * TILESIZE
        self.collision = [False] * 9

    def handle_keyboard_input(self):
        self.rot_speed = 0
        self.vel = vec(0, 0)
        keys = pygame.key.get_pressed()
        if keys[pg.K_LEFT] or keys[pg.K_a]:
            print("Left")
            self.rot_speed = PLAYER_ROT_SPEED
        if keys[pg.K_RIGHT] or keys[pg.K_d]:
            self.rot_speed = -PLAYER_ROT_SPEED
            print("Right")
        if keys[pg.K_UP] or keys[pg.K_w]:
            self.vel = vec(PLAYER_SPEED, 0).rotate(-self.rot)
            print("Up")
        if keys[pg.K_DOWN] or keys[pg.K_s]:
            self.vel = vec(-PLAYER_SPEED / 2, 0).rotate(-self.rot)
            print("down")

    def reset_pos(self):
        self.rect.y = random.randrange(350, 600)
        self.rect.x = random.randrange(0, self.screen_width[0])
        print(self.rect.y, self.rect.x)

    def update(self, pos, area, p1):
        """ Called each frame. """
        self.handle_keyboard_input()
        self.rot = (self.rot + self.rot_speed) % 360
        self.image = pg.transform.rotate(self.base_image, self.rot)
        self.image = pygame.transform.scale(self.image, (50, 50))
        self.rect = self.base_image.get_rect()
        self.rect.center = self.pos
        self.pos += self.vel
        # print(':::::::::',self.rect.x,self.rect.y)
        if self.vel[0] != 0:
            done_x = self.move_single_axis(self.vel[0], 0, area, p1)
        if self.vel[1] != 0:
            done_y = self.move_single_axis(0, self.vel[1], area, p1)

    def check_collision(self, rect):
        self.collision[0] = rect.collidepoint(self.rect.topleft)
        self.collision[1] = rect.collidepoint(self.rect.topright)
        self.collision[2] = rect.collidepoint(self.rect.bottomleft)
        self.collision[3] = rect.collidepoint(self.rect.bottomright)

        self.collision[4] = rect.collidepoint(self.rect.midleft)
        self.collision[5] = rect.collidepoint(self.rect.midright)
        self.collision[6] = rect.collidepoint(self.rect.midtop)
        self.collision[7] = rect.collidepoint(self.rect.midbottom)
        self.collision[8] = rect.collidepoint(self.rect.center)

        if True in self.collision:
            return self.collision.index(True)
        else:

            print(self.rect.topleft, self.rect.midtop, self.rect.midleft)

            for x in range(self.rect.topleft[0], self.rect.midtop[0]):
                if rect.collidepoint(x, self.rect.topleft[1]):
                    self.collision[0] = True
            for y in range(self.rect.topleft[1], self.rect.midleft[1]):
                if rect.collidepoint(self.rect.topleft[0], y):
                    self.collision[0] = True

            for x in range(self.rect.bottomleft[1], self.rect.midbottom[1]):
                if rect.collidepoint(x, self.rect.bottomleft[0]):
                    self.collision[2] = True
            for y in range(self.rect.bottomleft[0], self.rect.midleft[0]):
                if rect.collidepoint(self.rect.bottomleft[1], y):
                    self.collision[2] = True
            print("collision:", self.collision)
            return self.collision.index(True)

    def rotate(self, angle):
        self.image = pygame.transform.rotate(self.image, angle)
        self.rect = self.image.get_rect(center=self.rect.center)

    def move_single_axis(self, dx, dy, area, p1):
        # returns done

        # move ball rect
        self.rect.x += dx
        self.rect.y += dy

        if not area.contains(self.rect):
            # bottom wall
            if dy > 0:
                self.rect.bottom = area.bottom
                self.vel[1] = -self.vel[1]
            # top wall
            elif dy < 0:
                self.rect.top = area.top
                self.vel[1] = -self.vel[1]
            # right or left walls
            else:
                self.vel[0] = -self.vel[0]
                return True

        else:
            # Do ball and bat collide?
            # add some randomness
            r_val = 0
            if self.bounce_randomness:
                r_val = get_small_random_value()

            # ball in left half of screen
            is_collision, self.rect, self.vel = p1.process_collision(
                self.rect, dx, dy, self.vel
            )
            if is_collision:
                self.vel = vec(
                    0, 0
                )  # self.speed[0] + np.sign(self.speed[0]) * r_val, self.speed[1] + np.sign(self.speed[1]) * r_val]

        return False

    def process_collision(self, b_rect, dx, dy, b_speed):
        if self.rect.colliderect(b_rect):
            is_collision = True
            if dx < 0:
                b_rect.left = self.rect.right
                b_speed[0] = -b_speed[0]
            # top or bottom edge
            elif dy > 0:
                b_rect.bottom = self.rect.top
                b_speed[1] = -b_speed[1]
            elif dy < 0:
                b_rect.top = self.rect.bottom
                b_speed[1] = -b_speed[1]
            return is_collision, b_rect, b_speed
        return False, b_rect, b_speed


class agent2(pygame.sprite.Sprite):
    """
    This class represents the triangle
    """

    def __init__(self, _screen_width, speed=5):
        """ Constructor. Pass in the color of the block,
        and its x and y position. """
        # Call the parent class (Sprite) constructor
        super().__init__()

        self.image = get_image(
            "agent2.jpg"
        )  # pygame.draw.polygon(screen, BLACK, [[0, 0], [0, 50], [50, 0]], 5)#
        self.screen_width = _screen_width
        self.rect = self.image.get_rect()
        self.dim = self.rect.size
        self.rect.y = 350
        self.mode = True

        self.speed_val = speed
        self.vel = vec(
            int(self.speed_val * np.cos(np.pi / 4)),
            int(self.speed_val * np.sin(np.pi / 4)),
        )
        self.hit = False
        self.bounce_randomness = 1

    def reset_pos(self):
        self.rect.y = random.randrange(350, 600)
        self.rect.x = random.randrange(0, self.screen_width[0])
        # print(self.rect.y, self.rect.x)

    def update(self, pos):
        """ Called each frame. """
        if self.mode:
            # if self.rect.x > 1000:
            #     self.rect.x = 0
            # self.rect.x += 5
            if self.rect.y > 100:
                self.rect.center = pos
                # self.rect.x = pos[0]
            # else:
            #     self.reset_pos()

        else:
            if self.rect.y < 100:
                self.change_command()
                self.rect.y = 350
            self.rect.y -= 5

    def change_command(self):
        self.mode = not self.mode

    def process_collision(self, b_rect, dx, dy, b_speed):
        if self.rect.colliderect(b_rect):
            is_collision = True
            if dx < 0:
                b_rect.left = self.rect.right
                b_speed[0] = -b_speed[0]
            # top or bottom edge
            elif dy > 0:
                b_rect.bottom = self.rect.top
                b_speed[1] = -b_speed[1]
            elif dy < 0:
                b_rect.top = self.rect.bottom
                b_speed[1] = -b_speed[1]
            return is_collision, b_rect, b_speed
        return False, b_rect, b_speed

    def move_single_axis(self, dx, dy, area, p1):
        self.rect.x += dx
        self.rect.y += dy

        if not area.contains(self.rect):
            # bottom wall
            if dy > 0:
                self.rect.bottom = area.bottom
                self.vel[1] = -self.vel[1]
            # top wall
            elif dy < 0:
                self.rect.top = area.top
                self.vel[1] = -self.vel[1]
            # right or left walls
            else:
                self.vel[0] = -self.vel[0]
                return True

        else:
            r_val = 0
            if self.bounce_randomness:
                r_val = get_small_random_value()

            # ball in left half of screen
            is_collision, self.rect, self.vel = p1.process_collision(
                self.rect, dx, dy, self.vel
            )
            if is_collision:
                self.vel = vec(
                    0, 0
                )  # self.speed[0] + np.sign(self.speed[0]) * r_val, self.speed[1] + np.sign(self.speed[1]) * r_val]

        return False


class env(gym.Env):
    def __init__(self):
        super(env, self).__init__()
        global agent2, agent1
        pygame.init()

        # Set the width and height of the screen [width, height]
        size = (1002, 699)
        self.screen = pygame.display.set_mode(size)
        background = get_image("background.jpg")
        pygame.display.set_caption("My Game")
        self.screen.blit(background, (0, 0))
        self.area = self.screen.get_rect()

        # Loop until the user clicks the close button.
        done = False

        # Used to manage how fast the screen updates
        clock = pygame.time.Clock()
        agent1 = agent1(size, x=50, y=50, speed=20)
        agent2 = agent2(size)

        block_list, all_sprites_list = self.create_targets()

        vis = pygame.sprite.Group()  # Visualize block that is being carried by agent 1
        vis2 = pygame.sprite.Group()  # Visualize block that is being carried by agent 2
        block_picked = None
        block_transfered = None
        flag = 0
        blocks_hit_list = []
        # cropped = pygame.Surface((100,100))

        # -------- Main Program Loop -----------
        while not done:
            # --- Main event loop
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    done = True

            self.screen.blit(background, (0, 0))
            self.screen.blit(agent1.image, agent1.rect)
            # cropped.blit(agent1.image, (agent1.rect.x,agent1.rect.y))
            self.screen.blit(agent2.image, agent2.rect)
            pos = pygame.mouse.get_pos()
            agent1.update(pos, self.area, agent2)
            agent2.update(pos)  # , self.area, agent1)
            if flag == 0:
                blocks_hit_list = pygame.sprite.spritecollide(agent1, block_list, True)
            pygame.draw.circle(self.screen, RED, agent1.rect.topleft, 5)
            while len(blocks_hit_list) > 1:
                block_list.add(blocks_hit_list.pop())
            if blocks_hit_list:
                # print(len(blocks_hit_list), len(block_list))
                vis.add(blocks_hit_list[0])
                block_picked = blocks_hit_list[0]
                flag = 1
            if block_picked:
                corner = agent1.check_collision(block_picked.rect)

                block_picked.update(agent1.rect, corner)
                agent1.rotate_flag = True
            # --- Go ahead and update the screen with what we've drawn.
            # print(len(block_list))
            if agent1.rect.y < 355:
                flag = 0
                block_picked = None
                # agent1.rotate_flag = False

            blocks_transfer_list = pygame.sprite.spritecollide(agent2, vis, True)
            if blocks_transfer_list:
                block_transfered = blocks_transfer_list[0]
                block_transfered.update(agent2.rect)
                vis2.add(block_transfered)
                agent2.change_command()
                block_picked = None
                flag = 0

            if block_transfered:
                block_transfered.update(agent2.rect)
            if agent2.rect.y < 105:
                block_transfered = None

            block_list.draw(self.screen)
            vis.draw(self.screen)
            vis2.draw(self.screen)
            pygame.display.flip()
            clock.tick(10)

        pygame.quit()

    def create_targets(self):
        block_list = pygame.sprite.Group()

        # This is a list of every sprite.
        # All blocks and the player block as well.
        all_sprites_list = pygame.sprite.Group()
        x = 20
        for i in range(18):
            # This represents a block
            block = Block()
            # Set a random location for the block
            block.rect.x = x
            x += 75
            block.rect.y = 630
            # Add the block to the list of objects
            block_list.add(block)
            all_sprites_list.add(block)

        return block_list, all_sprites_list



from .multiwalker_base import MultiWalkerEnv as _env
from pettingzoo import AECEnv
from pettingzoo.utils import agent_selector
from pettingzoo.utils import EnvLogger
import numpy as np


class env(AECEnv):

    metadata = {'render.modes': ['human']}

    def __init__(self, seed=None, *args, **kwargs):
        super(env, self).__init__()
        self.env = _env(seed, *args, **kwargs)

        self.num_agents = self.env.num_agents
        self.agents = ["walker_" + str(r) for r in range(self.num_agents)]
        self.agent_name_mapping = dict(zip(self.agents, list(range(self.num_agents))))
        self.agent_order = self.agents[:]
        self._agent_selector = agent_selector(self.agent_order)
        # spaces
        self.action_spaces = dict(zip(self.agents, self.env.action_space))
        self.observation_spaces = dict(
            zip(self.agents, self.env.observation_space))
        self.steps = 0
        self.display_wait = 0.04
        self.observations = self.env.get_last_obs()

        self.has_reset = False

    def convert_to_dict(self, list_of_list):
        return dict(zip(self.agents, list_of_list))

    def reset(self, observe=True):
        self.has_reset = True
        self.env.reset()
        self.steps = 0
        self._agent_selector.reinit(self.agent_order)
        self.agent_selection = self._agent_selector.next()
        self.rewards = dict(
            zip(self.agents, [np.float64(0) for _ in self.agents]))
        self.dones = dict(zip(self.agents, [False for _ in self.agents]))
        self.infos = dict(zip(self.agents, [{} for _ in self.agents]))
        if observe:
            return self.observe(self.agent_selection)

    def close(self):
        if not self.has_reset:
            EnvLogger.warn_close_before_reset()
        else:
            self.env.close()

    def render(self, mode="human"):
        if not self.has_reset:
            EnvLogger.error_render_before_reset()
        else:
            self.env.render()

    def observe(self, agent):
        if not self.has_reset:
            EnvLogger.error_observe_before_reset()
        return self.env.observe(self.agent_name_mapping[agent])

    def step(self, action, observe=True):
        if not self.has_reset:
            EnvLogger.error_step_before_reset()
        agent = self.agent_selection
        if action is None or any(np.isnan(action)):
            EnvLogger.warn_action_is_NaN(backup_policy="setting to zeros")
            action = np.zeros_like(self.action_spaces[agent].sample())
        elif not self.action_spaces[agent].contains(action):
            EnvLogger.warn_action_out_of_bound(action=action, action_space=self.action_spaces[agent], backup_policy="setting to zeros")
            action = np.zeros_like(self.action_spaces[agent].sample())
        action = np.array(action, dtype=np.float32)
        self.env.step(action, self.agent_name_mapping[agent], self._agent_selector.is_last())
        for r in self.rewards:
            self.rewards[r] = self.env.get_last_rewards()[self.agent_name_mapping[r]]
        for d in self.dones:
            self.dones[d] = self.env.get_last_dones()[self.agent_name_mapping[r]]
        self.agent_selection = self._agent_selector.next()

        if self.env.frames >= self.env.max_frames:
            self.dones = dict(zip(self.agents, [True for _ in self.agents]))

        self.steps += 1
        if observe:
            return self.observe(self.agent_selection)



from ._mpe_utils.simple_env import SimpleEnv
from .scenarios.simple_reference import Scenario


class env(SimpleEnv):
    def __init__(self, seed=None, local_ratio=0.5, max_frames=100):
        assert 0. <= local_ratio <= 1., "local_ratio is a proportion. Must be between 0 and 1."
        scenario = Scenario()
        world = scenario.make_world()
        super(env, self).__init__(scenario, world, max_frames, seed, local_ratio)



import gym

if __name__ == '__main__':
    env = gym.make('gym_backgammon:backgammon-v0')
    assert env.observation_space.low[0] == 0.0 and env.observation_space.high[0] == 1.0
    assert env.observation_space.low[1] == 0.0 and env.observation_space.high[1] == 1.0
    assert env.observation_space.low[2] == 0.0 and env.observation_space.high[2] == 1.0
    assert env.observation_space.low[3] == 0.0 and env.observation_space.high[3] == 6.0
    assert env.observation_space.low[4] == 0.0 and env.observation_space.high[4] == 1.0
    assert env.observation_space.low[5] == 0.0 and env.observation_space.high[5] == 1.0
    assert env.observation_space.low[6] == 0.0 and env.observation_space.high[6] == 1.0
    assert env.observation_space.low[7] == 0.0 and env.observation_space.high[7] == 6.0
    assert env.observation_space.low[8] == 0.0 and env.observation_space.high[8] == 1.0
    assert env.observation_space.low[9] == 0.0 and env.observation_space.high[9] == 1.0
    assert env.observation_space.low[10] == 0.0 and env.observation_space.high[10] == 1.0
    assert env.observation_space.low[11] == 0.0 and env.observation_space.high[11] == 6.0
    assert env.observation_space.low[12] == 0.0 and env.observation_space.high[12] == 1.0
    assert env.observation_space.low[13] == 0.0 and env.observation_space.high[13] == 1.0
    assert env.observation_space.low[14] == 0.0 and env.observation_space.high[14] == 1.0
    assert env.observation_space.low[15] == 0.0 and env.observation_space.high[15] == 6.0
    assert env.observation_space.low[16] == 0.0 and env.observation_space.high[16] == 1.0
    assert env.observation_space.low[17] == 0.0 and env.observation_space.high[17] == 1.0
    assert env.observation_space.low[18] == 0.0 and env.observation_space.high[18] == 1.0
    assert env.observation_space.low[19] == 0.0 and env.observation_space.high[19] == 6.0
    assert env.observation_space.low[20] == 0.0 and env.observation_space.high[20] == 1.0
    assert env.observation_space.low[21] == 0.0 and env.observation_space.high[21] == 1.0
    assert env.observation_space.low[22] == 0.0 and env.observation_space.high[22] == 1.0
    assert env.observation_space.low[23] == 0.0 and env.observation_space.high[23] == 6.0
    assert env.observation_space.low[24] == 0.0 and env.observation_space.high[24] == 1.0
    assert env.observation_space.low[25] == 0.0 and env.observation_space.high[25] == 1.0
    assert env.observation_space.low[26] == 0.0 and env.observation_space.high[26] == 1.0
    assert env.observation_space.low[27] == 0.0 and env.observation_space.high[27] == 6.0
    assert env.observation_space.low[28] == 0.0 and env.observation_space.high[28] == 1.0
    assert env.observation_space.low[29] == 0.0 and env.observation_space.high[29] == 1.0
    assert env.observation_space.low[30] == 0.0 and env.observation_space.high[30] == 1.0
    assert env.observation_space.low[31] == 0.0 and env.observation_space.high[31] == 6.0
    assert env.observation_space.low[32] == 0.0 and env.observation_space.high[32] == 1.0
    assert env.observation_space.low[33] == 0.0 and env.observation_space.high[33] == 1.0
    assert env.observation_space.low[34] == 0.0 and env.observation_space.high[34] == 1.0
    assert env.observation_space.low[35] == 0.0 and env.observation_space.high[35] == 6.0
    assert env.observation_space.low[36] == 0.0 and env.observation_space.high[36] == 1.0
    assert env.observation_space.low[37] == 0.0 and env.observation_space.high[37] == 1.0
    assert env.observation_space.low[38] == 0.0 and env.observation_space.high[38] == 1.0
    assert env.observation_space.low[39] == 0.0 and env.observation_space.high[39] == 6.0
    assert env.observation_space.low[40] == 0.0 and env.observation_space.high[40] == 1.0
    assert env.observation_space.low[41] == 0.0 and env.observation_space.high[41] == 1.0
    assert env.observation_space.low[42] == 0.0 and env.observation_space.high[42] == 1.0
    assert env.observation_space.low[43] == 0.0 and env.observation_space.high[43] == 6.0
    assert env.observation_space.low[44] == 0.0 and env.observation_space.high[44] == 1.0
    assert env.observation_space.low[45] == 0.0 and env.observation_space.high[45] == 1.0
    assert env.observation_space.low[46] == 0.0 and env.observation_space.high[46] == 1.0
    assert env.observation_space.low[47] == 0.0 and env.observation_space.high[47] == 6.0
    assert env.observation_space.low[48] == 0.0 and env.observation_space.high[48] == 1.0
    assert env.observation_space.low[49] == 0.0 and env.observation_space.high[49] == 1.0
    assert env.observation_space.low[50] == 0.0 and env.observation_space.high[50] == 1.0
    assert env.observation_space.low[51] == 0.0 and env.observation_space.high[51] == 6.0
    assert env.observation_space.low[52] == 0.0 and env.observation_space.high[52] == 1.0
    assert env.observation_space.low[53] == 0.0 and env.observation_space.high[53] == 1.0
    assert env.observation_space.low[54] == 0.0 and env.observation_space.high[54] == 1.0
    assert env.observation_space.low[55] == 0.0 and env.observation_space.high[55] == 6.0
    assert env.observation_space.low[56] == 0.0 and env.observation_space.high[56] == 1.0
    assert env.observation_space.low[57] == 0.0 and env.observation_space.high[57] == 1.0
    assert env.observation_space.low[58] == 0.0 and env.observation_space.high[58] == 1.0
    assert env.observation_space.low[59] == 0.0 and env.observation_space.high[59] == 6.0
    assert env.observation_space.low[60] == 0.0 and env.observation_space.high[60] == 1.0
    assert env.observation_space.low[61] == 0.0 and env.observation_space.high[61] == 1.0
    assert env.observation_space.low[62] == 0.0 and env.observation_space.high[62] == 1.0
    assert env.observation_space.low[63] == 0.0 and env.observation_space.high[63] == 6.0
    assert env.observation_space.low[64] == 0.0 and env.observation_space.high[64] == 1.0
    assert env.observation_space.low[65] == 0.0 and env.observation_space.high[65] == 1.0
    assert env.observation_space.low[66] == 0.0 and env.observation_space.high[66] == 1.0
    assert env.observation_space.low[67] == 0.0 and env.observation_space.high[67] == 6.0
    assert env.observation_space.low[68] == 0.0 and env.observation_space.high[68] == 1.0
    assert env.observation_space.low[69] == 0.0 and env.observation_space.high[69] == 1.0
    assert env.observation_space.low[70] == 0.0 and env.observation_space.high[70] == 1.0
    assert env.observation_space.low[71] == 0.0 and env.observation_space.high[71] == 6.0
    assert env.observation_space.low[72] == 0.0 and env.observation_space.high[72] == 1.0
    assert env.observation_space.low[73] == 0.0 and env.observation_space.high[73] == 1.0
    assert env.observation_space.low[74] == 0.0 and env.observation_space.high[74] == 1.0
    assert env.observation_space.low[75] == 0.0 and env.observation_space.high[75] == 6.0
    assert env.observation_space.low[76] == 0.0 and env.observation_space.high[76] == 1.0
    assert env.observation_space.low[77] == 0.0 and env.observation_space.high[77] == 1.0
    assert env.observation_space.low[78] == 0.0 and env.observation_space.high[78] == 1.0
    assert env.observation_space.low[79] == 0.0 and env.observation_space.high[79] == 6.0
    assert env.observation_space.low[80] == 0.0 and env.observation_space.high[80] == 1.0
    assert env.observation_space.low[81] == 0.0 and env.observation_space.high[81] == 1.0
    assert env.observation_space.low[82] == 0.0 and env.observation_space.high[82] == 1.0
    assert env.observation_space.low[83] == 0.0 and env.observation_space.high[83] == 6.0
    assert env.observation_space.low[84] == 0.0 and env.observation_space.high[84] == 1.0
    assert env.observation_space.low[85] == 0.0 and env.observation_space.high[85] == 1.0
    assert env.observation_space.low[86] == 0.0 and env.observation_space.high[86] == 1.0
    assert env.observation_space.low[87] == 0.0 and env.observation_space.high[87] == 6.0
    assert env.observation_space.low[88] == 0.0 and env.observation_space.high[88] == 1.0
    assert env.observation_space.low[89] == 0.0 and env.observation_space.high[89] == 1.0
    assert env.observation_space.low[90] == 0.0 and env.observation_space.high[90] == 1.0
    assert env.observation_space.low[91] == 0.0 and env.observation_space.high[91] == 6.0
    assert env.observation_space.low[92] == 0.0 and env.observation_space.high[92] == 1.0
    assert env.observation_space.low[93] == 0.0 and env.observation_space.high[93] == 1.0
    assert env.observation_space.low[94] == 0.0 and env.observation_space.high[94] == 1.0
    assert env.observation_space.low[95] == 0.0 and env.observation_space.high[95] == 6.0
    assert env.observation_space.low[96] == 0.0 and env.observation_space.high[96] == 7.5
    assert env.observation_space.low[97] == 0.0 and env.observation_space.high[97] == 1.0
    assert env.observation_space.low[98] == 0.0 and env.observation_space.high[98] == 1.0
    assert env.observation_space.low[99] == 0.0 and env.observation_space.high[99] == 1.0
    assert env.observation_space.low[100] == 0.0 and env.observation_space.high[100] == 1.0
    assert env.observation_space.low[101] == 0.0 and env.observation_space.high[101] == 6.0
    assert env.observation_space.low[102] == 0.0 and env.observation_space.high[102] == 1.0
    assert env.observation_space.low[103] == 0.0 and env.observation_space.high[103] == 1.0
    assert env.observation_space.low[104] == 0.0 and env.observation_space.high[104] == 1.0
    assert env.observation_space.low[105] == 0.0 and env.observation_space.high[105] == 6.0
    assert env.observation_space.low[106] == 0.0 and env.observation_space.high[106] == 1.0
    assert env.observation_space.low[107] == 0.0 and env.observation_space.high[107] == 1.0
    assert env.observation_space.low[108] == 0.0 and env.observation_space.high[108] == 1.0
    assert env.observation_space.low[109] == 0.0 and env.observation_space.high[109] == 6.0
    assert env.observation_space.low[110] == 0.0 and env.observation_space.high[110] == 1.0
    assert env.observation_space.low[111] == 0.0 and env.observation_space.high[111] == 1.0
    assert env.observation_space.low[112] == 0.0 and env.observation_space.high[112] == 1.0
    assert env.observation_space.low[113] == 0.0 and env.observation_space.high[113] == 6.0
    assert env.observation_space.low[114] == 0.0 and env.observation_space.high[114] == 1.0
    assert env.observation_space.low[115] == 0.0 and env.observation_space.high[115] == 1.0
    assert env.observation_space.low[116] == 0.0 and env.observation_space.high[116] == 1.0
    assert env.observation_space.low[117] == 0.0 and env.observation_space.high[117] == 6.0
    assert env.observation_space.low[118] == 0.0 and env.observation_space.high[118] == 1.0
    assert env.observation_space.low[119] == 0.0 and env.observation_space.high[119] == 1.0
    assert env.observation_space.low[120] == 0.0 and env.observation_space.high[120] == 1.0
    assert env.observation_space.low[121] == 0.0 and env.observation_space.high[121] == 6.0
    assert env.observation_space.low[122] == 0.0 and env.observation_space.high[122] == 1.0
    assert env.observation_space.low[123] == 0.0 and env.observation_space.high[123] == 1.0
    assert env.observation_space.low[124] == 0.0 and env.observation_space.high[124] == 1.0
    assert env.observation_space.low[125] == 0.0 and env.observation_space.high[125] == 6.0
    assert env.observation_space.low[126] == 0.0 and env.observation_space.high[126] == 1.0
    assert env.observation_space.low[127] == 0.0 and env.observation_space.high[127] == 1.0
    assert env.observation_space.low[128] == 0.0 and env.observation_space.high[128] == 1.0
    assert env.observation_space.low[129] == 0.0 and env.observation_space.high[129] == 6.0
    assert env.observation_space.low[130] == 0.0 and env.observation_space.high[130] == 1.0
    assert env.observation_space.low[131] == 0.0 and env.observation_space.high[131] == 1.0
    assert env.observation_space.low[132] == 0.0 and env.observation_space.high[132] == 1.0
    assert env.observation_space.low[133] == 0.0 and env.observation_space.high[133] == 6.0
    assert env.observation_space.low[134] == 0.0 and env.observation_space.high[134] == 1.0
    assert env.observation_space.low[135] == 0.0 and env.observation_space.high[135] == 1.0
    assert env.observation_space.low[136] == 0.0 and env.observation_space.high[136] == 1.0
    assert env.observation_space.low[137] == 0.0 and env.observation_space.high[137] == 6.0
    assert env.observation_space.low[138] == 0.0 and env.observation_space.high[138] == 1.0
    assert env.observation_space.low[139] == 0.0 and env.observation_space.high[139] == 1.0
    assert env.observation_space.low[140] == 0.0 and env.observation_space.high[140] == 1.0
    assert env.observation_space.low[141] == 0.0 and env.observation_space.high[141] == 6.0
    assert env.observation_space.low[142] == 0.0 and env.observation_space.high[142] == 1.0
    assert env.observation_space.low[143] == 0.0 and env.observation_space.high[143] == 1.0
    assert env.observation_space.low[144] == 0.0 and env.observation_space.high[144] == 1.0
    assert env.observation_space.low[145] == 0.0 and env.observation_space.high[145] == 6.0
    assert env.observation_space.low[146] == 0.0 and env.observation_space.high[146] == 1.0
    assert env.observation_space.low[147] == 0.0 and env.observation_space.high[147] == 1.0
    assert env.observation_space.low[148] == 0.0 and env.observation_space.high[148] == 1.0
    assert env.observation_space.low[149] == 0.0 and env.observation_space.high[149] == 6.0
    assert env.observation_space.low[150] == 0.0 and env.observation_space.high[150] == 1.0
    assert env.observation_space.low[151] == 0.0 and env.observation_space.high[151] == 1.0
    assert env.observation_space.low[152] == 0.0 and env.observation_space.high[152] == 1.0
    assert env.observation_space.low[153] == 0.0 and env.observation_space.high[153] == 6.0
    assert env.observation_space.low[154] == 0.0 and env.observation_space.high[154] == 1.0
    assert env.observation_space.low[155] == 0.0 and env.observation_space.high[155] == 1.0
    assert env.observation_space.low[156] == 0.0 and env.observation_space.high[156] == 1.0
    assert env.observation_space.low[157] == 0.0 and env.observation_space.high[157] == 6.0
    assert env.observation_space.low[158] == 0.0 and env.observation_space.high[158] == 1.0
    assert env.observation_space.low[159] == 0.0 and env.observation_space.high[159] == 1.0
    assert env.observation_space.low[160] == 0.0 and env.observation_space.high[160] == 1.0
    assert env.observation_space.low[161] == 0.0 and env.observation_space.high[161] == 6.0
    assert env.observation_space.low[162] == 0.0 and env.observation_space.high[162] == 1.0
    assert env.observation_space.low[163] == 0.0 and env.observation_space.high[163] == 1.0
    assert env.observation_space.low[164] == 0.0 and env.observation_space.high[164] == 1.0
    assert env.observation_space.low[165] == 0.0 and env.observation_space.high[165] == 6.0
    assert env.observation_space.low[166] == 0.0 and env.observation_space.high[166] == 1.0
    assert env.observation_space.low[167] == 0.0 and env.observation_space.high[167] == 1.0
    assert env.observation_space.low[168] == 0.0 and env.observation_space.high[168] == 1.0
    assert env.observation_space.low[169] == 0.0 and env.observation_space.high[169] == 6.0
    assert env.observation_space.low[170] == 0.0 and env.observation_space.high[170] == 1.0
    assert env.observation_space.low[171] == 0.0 and env.observation_space.high[171] == 1.0
    assert env.observation_space.low[172] == 0.0 and env.observation_space.high[172] == 1.0
    assert env.observation_space.low[173] == 0.0 and env.observation_space.high[173] == 6.0
    assert env.observation_space.low[174] == 0.0 and env.observation_space.high[174] == 1.0
    assert env.observation_space.low[175] == 0.0 and env.observation_space.high[175] == 1.0
    assert env.observation_space.low[176] == 0.0 and env.observation_space.high[176] == 1.0
    assert env.observation_space.low[177] == 0.0 and env.observation_space.high[177] == 6.0
    assert env.observation_space.low[178] == 0.0 and env.observation_space.high[178] == 1.0
    assert env.observation_space.low[179] == 0.0 and env.observation_space.high[179] == 1.0
    assert env.observation_space.low[180] == 0.0 and env.observation_space.high[180] == 1.0
    assert env.observation_space.low[181] == 0.0 and env.observation_space.high[181] == 6.0
    assert env.observation_space.low[182] == 0.0 and env.observation_space.high[182] == 1.0
    assert env.observation_space.low[183] == 0.0 and env.observation_space.high[183] == 1.0
    assert env.observation_space.low[184] == 0.0 and env.observation_space.high[184] == 1.0
    assert env.observation_space.low[185] == 0.0 and env.observation_space.high[185] == 6.0
    assert env.observation_space.low[186] == 0.0 and env.observation_space.high[186] == 1.0
    assert env.observation_space.low[187] == 0.0 and env.observation_space.high[187] == 1.0
    assert env.observation_space.low[188] == 0.0 and env.observation_space.high[188] == 1.0
    assert env.observation_space.low[189] == 0.0 and env.observation_space.high[189] == 6.0
    assert env.observation_space.low[190] == 0.0 and env.observation_space.high[190] == 1.0
    assert env.observation_space.low[191] == 0.0 and env.observation_space.high[191] == 1.0
    assert env.observation_space.low[192] == 0.0 and env.observation_space.high[192] == 1.0
    assert env.observation_space.low[193] == 0.0 and env.observation_space.high[193] == 6.0
    assert env.observation_space.low[194] == 0.0 and env.observation_space.high[194] == 7.5
    assert env.observation_space.low[195] == 0.0 and env.observation_space.high[195] == 1.0
    assert env.observation_space.low[196] == 0.0 and env.observation_space.high[196] == 1.0
    assert env.observation_space.low[197] == 0.0 and env.observation_space.high[197] == 1.0



import pygame as pg
import pygame.locals as locals
import pymunk as pm
from pymunk import Vec2d
from gym import spaces
import numpy as np

from pettingzoo import AECEnv
from pettingzoo.utils import agent_selector
import pettingzoo.gamma.prospector.constants as const
import pettingzoo.gamma.prospector.utils as utils

import math
import os
from enum import IntEnum, auto
import itertools as it
from random import randint


class CollisionTypes(IntEnum):
    PROSPECTOR = auto()
    BOUNDARY = auto()
    WATER = auto()
    CHEST = auto()
    GOLD = auto()
    BANKER = auto()


class Prospector(pg.sprite.Sprite):
    def __init__(self, pos, space, *sprite_groups):
        super().__init__(sprite_groups)
        # self.image = load_image(['prospec.png'])
        self.image = utils.load_image(["prospector-pickaxe-big.png"])
        self.image = pg.transform.scale(
            self.image, (int(const.AGENT_RADIUS * 2), int(const.AGENT_RADIUS * 2))
        )

        self.rect = self.image.get_rect()
        self.orig_image = self.image

        # Create the physics body and shape of this object.
        # moment = pm.moment_for_poly(mass, vertices)

        moment = pm.moment_for_circle(1, 0, self.rect.width / 2)

        self.body = pm.Body(1, moment)
        self.body.nugget = None
        self.body.sprite_type = "prospector"
        # self.shape = pm.Poly(self.body, vertices, radius=3)

        self.shape = pm.Circle(self.body, const.AGENT_RADIUS)
        self.shape.elasticity = 0.0
        self.shape.collision_type = CollisionTypes.PROSPECTOR

        self.body.position = utils.flipy(pos)
        # Add them to the Pymunk space.
        self.space = space
        self.space.add(self.body, self.shape)

    @property
    def center(self):
        return self.rect.x + const.AGENT_RADIUS, self.rect.y + const.AGENT_RADIUS

    def _update(self, action):
        # forward/backward action
        y_vel = action[0] * const.PROSPECTOR_SPEED
        # left/right action
        x_vel = action[1] * const.PROSPECTOR_SPEED

        delta_angle = action[2] * const.MAX_SPRITE_ROTATION

        self.body.angle += delta_angle
        self.body.angular_velocity = 0

        self.body.velocity = Vec2d(x_vel, y_vel).rotated(self.body.angle)

        self.rect.center = utils.flipy(self.body.position)
        self.image = pg.transform.rotozoom(
            self.orig_image, math.degrees(self.body.angle), 1
        )
        self.rect = self.image.get_rect(center=self.rect.center)

        curr_vel = self.body.velocity

        if self.body.nugget is not None:
            self.body.nugget.update(self.body.position, self.body.angle, False)

        self.body.velocity = curr_vel

    def update(self, keys):
        x = y = 0
        if keys[locals.K_w]:
            y = 1
        if keys[locals.K_s]:
            y = -1
        if keys[locals.K_d]:
            x = 1
        if keys[locals.K_a]:
            x = -1
        if keys[locals.K_q]:
            self.body.angle += 0.1
            self.body.angular_velocity = 0
        if keys[locals.K_e]:
            self.body.angle -= 0.1
            self.body.angular_velocity = 0

        if x != 0 and y != 0:
            self.body.velocity = Vec2d(x, y).rotated(self.body.angle) * (
                const.PROSPECTOR_SPEED / math.sqrt(2)
            )
        else:
            self.body.velocity = (
                Vec2d(x, y).rotated(self.body.angle) * const.PROSPECTOR_SPEED
            )

        # Rotate the image of the sprite.
        self.rect.center = utils.flipy(self.body.position)
        self.image = pg.transform.rotozoom(
            self.orig_image, math.degrees(self.body.angle), 1
        )
        self.rect = self.image.get_rect(center=self.rect.center)

        curr_vel = self.body.velocity

        if self.body.nugget is not None:
            self.body.nugget.update(self.body.position, self.body.angle, False)

        self.body.velocity = curr_vel


class Banker(pg.sprite.Sprite):
    def __init__(self, pos, space, num, *sprite_groups):
        super().__init__(sprite_groups)
        self.image = utils.load_image(["bankers", "%s-big.png" % num])
        self.image = pg.transform.scale(
            self.image, (int(const.AGENT_RADIUS * 2), int(const.AGENT_RADIUS * 2))
        )

        self.rect = self.image.get_rect()
        self.orig_image = self.image

        moment = pm.moment_for_circle(1, 0, self.rect.width / 2)

        self.body = pm.Body(1, moment)
        self.body.nugget = None
        self.body.sprite_type = "banker"
        self.body.nugget_offset = None

        self.shape = pm.Circle(self.body, const.AGENT_RADIUS)
        self.shape.collision_type = CollisionTypes.BANKER

        self.body.position = utils.flipy(pos)
        # Add them to the Pymunk space.
        self.space = space
        self.space.add(self.body, self.shape)

    @property
    def center(self):
        return self.rect.x + const.AGENT_RADIUS, self.rect.y + const.AGENT_RADIUS

    def update(self, keys):
        move = 0
        if any(
            keys[key]
            for key in (locals.K_UP, locals.K_DOWN, locals.K_RIGHT, locals.K_LEFT,)
        ):
            move = 1

        if keys[locals.K_UP]:
            self.body.angle = 0
        elif keys[locals.K_DOWN]:
            self.body.angle = math.pi
        elif keys[locals.K_RIGHT]:
            self.body.angle = -math.pi / 2
        elif keys[locals.K_LEFT]:
            self.body.angle = math.pi / 2

        self.body.velocity = (
            Vec2d(0, move).rotated(self.body.angle) * const.BANKER_SPEED
        )

        # Rotate the image of the sprite.
        self.rect.center = utils.flipy(self.body.position)
        self.image = pg.transform.rotozoom(
            self.orig_image, math.degrees(self.body.angle), 1
        )

        self.rect = self.image.get_rect(center=self.rect.center)

        curr_vel = self.body.velocity

        if self.body.nugget is not None:
            corrected = utils.normalize_angle(self.body.angle + (math.pi / 2))
            self.body.nugget.update(self.body.position, corrected, True)

        self.body.velocity = curr_vel

    def _update(self, action):
        # left/right action
        x_vel = action[0] * const.PROSPECTOR_SPEED
        # up/down action
        y_vel = action[1] * const.PROSPECTOR_SPEED

        # Add math.pi / 2 because sprite is facing upwards at the start
        angle_radians = math.atan2(x_vel, y_vel) + (math.pi / 2)

        # Angle is determined only by current trajectory.
        self.body.angle = angle_radians
        self.body.angular_velocity = 0

        self.body.velocity = Vec2d(x_vel, y_vel).rotated(self.body.angle)

        self.rect.center = utils.flipy(self.body.position)
        self.image = pg.transform.rotozoom(
            self.orig_image, math.degrees(self.body.angle), 1
        )
        self.rect = self.image.get_rect(center=self.rect.center)

        curr_vel = self.body.velocity

        if self.body.nugget is not None:
            self.body.nugget.update(self.body.position, self.body.angle, False)

        self.body.velocity = curr_vel


class Fence(pg.sprite.Sprite):
    def __init__(self, w_type, sprite_pos, body_pos, verts, space, *sprite_groups):
        super().__init__(sprite_groups)

        if w_type == "top":
            # self.image = utils.load_image(["horiz-fence.png"])
            self.image = utils.load_image(["top-down-horiz-fence.png"])
        elif w_type in ["right", "left"]:
            # self.image = utils.load_image(["vert-fence.png"])
            self.image = utils.load_image(["top-down-vert-fence.png"])
        else:
            raise ValueError("Fence image not found! Check the spelling")
        # elif w_type == "left":
        #     # self.image = utils.load_image(["vert-fence.png"])
        #     self.image = utils.load_image(["top-down-vert-fence.png"])

        self.rect = self.image.get_rect(topleft=sprite_pos)

        self.body = pm.Body(body_type=pm.Body.STATIC)

        # Transform pygame vertices to fit Pymunk body
        invert_verts = utils.invert_y(verts)
        self.shape = pm.Poly(self.body, invert_verts)
        self.shape.elasticity = 0.0
        self.shape.collision_type = CollisionTypes.BOUNDARY

        self.body.position = utils.flipy(body_pos)
        space.add(self.shape)


class Water(pg.sprite.Sprite):
    def __init__(self, pos, verts, space, *sprite_groups):
        super().__init__(*sprite_groups)
        # Determine the width and height of the surface.
        self.image = utils.load_image(["water.png"])
        self.image = pg.transform.scale(
            self.image, (const.SCREEN_WIDTH, const.WATER_HEIGHT)
        )

        self.rect = self.image.get_rect(topleft=pos)

        self.body = pm.Body(body_type=pm.Body.STATIC)

        # Transform pygame vertices to fit Pymunk body
        invert_verts = utils.invert_y(verts)
        self.shape = pm.Poly(self.body, invert_verts)
        self.shape.collision_type = CollisionTypes.WATER

        # self.shape.friction = 1.0
        self.body.position = utils.flipy(pos)
        self.space = space
        self.space.add(self.shape)


class Bank(pg.sprite.Sprite):
    def __init__(self, pos, verts, space, *sprite_groups):
        super().__init__(sprite_groups)

        self.image = utils.load_image(["bank-2.png"])
        self.image = pg.transform.scale(self.image, (184, 100))
        self.rect = self.image.get_rect(topleft=pos)

        self.body = pm.Body(body_type=pm.Body.STATIC)
        self.body.score = 0

        invert_verts = utils.invert_y(verts)
        self.shape = pm.Poly(self.body, invert_verts)
        self.shape.collision_type = CollisionTypes.CHEST

        self.body.position = utils.flipy(pos)
        self.space = space
        self.space.add(self.shape, self.body)

    def __str__(self):
        return str(self.body.score)


class Gold(pg.sprite.Sprite):
    ids = it.count(0)

    def __init__(self, pos, body, space, *sprite_groups):
        super().__init__(sprite_groups)
        self.id = next(self.ids)

        self.image = utils.load_image(["gold", "6.png"])
        self.image = pg.transform.scale(self.image, (16, 16))
        self.orig_image = self.image

        self.rect = self.image.get_rect()

        self.moment = pm.moment_for_circle(1, 0, 8)
        self.body = pm.Body(1, self.moment)
        self.body.position = body.position

        self.shape = pm.Circle(self.body, 8)
        self.shape.collision_type = CollisionTypes.GOLD
        self.shape.id = self.id

        self.space = space
        self.space.add(self.body, self.shape)

        self.initial_angle = body.angle - Vec2d(0, -1).angle
        self.parent_body = body

    def update(self, pos, angle, banker: bool):

        if banker:
            new_angle = angle
        else:
            new_angle = angle - self.initial_angle
        new_pos = pos + Vec2d(const.AGENT_RADIUS + 9, 0).rotated(new_angle)

        self.body.position = new_pos
        self.body.angular_velocity = 0
        self.rect.center = utils.flipy(self.body.position)
        self.image = pg.transform.rotozoom(
            self.orig_image, math.degrees(self.body.angle), 1
        )
        self.rect = self.image.get_rect(center=self.rect.center)

    def draw(self, surf):
        surf.blit(self.image, self.rect)


class Game:
    def __init__(self):
        self.done = False
        self.screen = pg.display.set_mode(const.SCREEN_SIZE)
        self.clock = pg.time.Clock()

        self.background = utils.load_image(["background-debris.png"])
        # self.background_rect = pg.Rect(0, 0, *const.SCREEN_SIZE)
        self.background_rect = self.background.get_rect(topleft=(0, 0))

        self.space = pm.Space()
        self.space.gravity = Vec2d(0.0, 0.0)
        # self.space.damping = 0.5
        self.space.damping = 0.0

        self.all_sprites = pg.sprite.Group()
        self.gold = []

        prospec_info = [utils.rand_pos("prospector") for _ in range(3)]
        self.prospectors = []
        for pos in prospec_info:
            self.prospectors.append(Prospector(pos, self.space, self.all_sprites))

        banker_info = (
            (1, utils.rand_pos("banker")),
            (2, utils.rand_pos("banker")),
            (3, utils.rand_pos("banker")),
        )
        self.bankers = []
        for num, pos in banker_info:
            self.bankers.append(Banker(pos, self.space, num, self.all_sprites))

        chest_verts = (
            (0, 0),
            (184, 0),
            (184, 100),
            (0, 100),
        )

        chest_info = [
            ([184 * 1, 50], chest_verts),
            ([184 * 3, 50], chest_verts),
            ([184 * 5, 50], chest_verts),
        ]

        self.chests = []
        for pos, verts in chest_info:
            self.chests.append(Bank(pos, verts, self.space, self.all_sprites))

        for w_type, s_pos, b_pos, verts in const.FENCE_INFO:
            Fence(w_type, s_pos, b_pos, verts, self.space, self.all_sprites)

        water_info = {
            "pos": (0, const.SCREEN_HEIGHT - const.WATER_HEIGHT),
            "verts": (
                (0, 0),
                (const.SCREEN_WIDTH, 0),
                (const.SCREEN_WIDTH, const.WATER_HEIGHT),
                (0, const.WATER_HEIGHT),
            ),
        }

        Water(water_info["pos"], water_info["verts"], self.space, self.all_sprites)

        def add_gold(arbiter, space, data):
            prospec = arbiter.shapes[0]
            prospec_body = prospec.body

            position = arbiter.contact_point_set.points[0].point_a
            normal = arbiter.contact_point_set.normal

            prospec_body.position = position - (24 * normal)
            prospec_body.velocity = (0, 0)

            if prospec_body.nugget is None:

                position = arbiter.contact_point_set.points[0].point_a

                gold = Gold(position, prospec_body, self.space)
                self.gold.append(gold)

                prospec_body.nugget = gold

            return True

        gold_dispenser = self.space.add_collision_handler(
            CollisionTypes.PROSPECTOR, CollisionTypes.WATER
        )

        gold_dispenser.begin = add_gold

        def handoff_gold_handler(arbiter, space, data):
            banker, gold = arbiter.shapes[0], arbiter.shapes[1]

            gold_class = None
            for g in self.gold:
                if g.id == gold.id:
                    gold_class = g

            if gold_class.parent_body.sprite_type == "prospector":

                banker_body = banker.body

                normal = arbiter.contact_point_set.normal

                corrected = utils.normalize_angle(banker_body.angle + (math.pi / 2))

                if (
                    corrected - const.BANKER_HANDOFF_TOLERANCE
                    <= normal.angle
                    <= corrected + const.BANKER_HANDOFF_TOLERANCE
                ):

                    gold_class.parent_body.nugget = None

                    gold_class.parent_body = banker_body
                    banker_body.nugget = gold_class
                    banker_body.nugget_offset = normal.angle

            return True

        handoff_gold = self.space.add_collision_handler(
            CollisionTypes.BANKER, CollisionTypes.GOLD
        )

        handoff_gold.begin = handoff_gold_handler

        def gold_score_handler(arbiter, space, data):
            gold, chest = arbiter.shapes[0], arbiter.shapes[1]

            gold_class = None
            for g in self.gold:
                if g.id == gold.id:
                    gold_class = g

            if gold_class.parent_body.sprite_type == "banker":

                chest.body.score += 1

                self.space.remove(gold, gold.body)

                gold_class.parent_body.nugget = None

                # total_score = ", ".join(
                #     [
                #         "Chest %d: %d" % (i, c.body.score)
                #         for i, c in enumerate(self.chests)
                #     ]
                # )

                # print(total_score)

                self.gold.remove(gold_class)
                self.all_sprites.remove(gold_class)

            return False

        gold_score = self.space.add_collision_handler(
            CollisionTypes.GOLD, CollisionTypes.CHEST
        )

        gold_score.begin = gold_score_handler

    def run(self):
        while not self.done:
            for event in pg.event.get():
                if event.type == locals.QUIT or (
                    event.type == locals.KEYDOWN and event.key in [locals.K_ESCAPE]
                ):
                    self.done = True

            self.dt = self.clock.tick(15)

            self.space.step(1 / 15)
            self.all_sprites.update(pg.key.get_pressed())

            self.draw()

    def draw(self):
        # self.screen.fill(BACKGROUND_COLOR)
        # self.background.blit(self.screen)
        self.screen.blit(self.background, self.background_rect)
        self.all_sprites.draw(self.screen)

        for p in self.prospectors:
            if p.body.nugget is not None:
                p.body.nugget.draw(self.screen)

        for b in self.bankers:
            if b.body.nugget is not None:
                b.body.nugget.draw(self.screen)

        pg.display.flip()


class env(AECEnv):
    def __init__(
        self,
        ind_reward=0.8,
        group_reward=0.1,
        other_group_reward=0.1,
        prospec_find_gold_reward=1,
        prospec_handoff_gold_reward=1,
        banker_receive_gold_reward=1,
        banker_deposit_gold_reward=1,
        max_frames=900,
    ):
        if ind_reward + group_reward + other_group_reward != 1.0:
            raise ValueError(
                "Individual reward, group reward, and other group reward should "
                "add up to 1.0"
            )

        self.num_agents = 7
        # self.agents = list(range(0, self.num_agents))
        self.agents = []

        self.sprite_list = [
            "bankers/1-big.png",
            "bankers/2-big.png",
            "bankers/3-big.png",
            "prospector-pickaxe-big.png",
        ]
        self.rendering = False
        self.max_frames = max_frames
        self.frame = 0

        # TODO: Setup game data here
        pg.init()
        self.clock = pg.time.Clock()
        self.done = False

        self.background = utils.load_image(["test.png"])
        self.background_rect = pg.Rect(0, 0, *const.SCREEN_SIZE)

        self.space = pm.Space()
        self.space.gravity = Vec2d(0.0, 0.0)
        self.space.damping = 0.0

        self.all_sprites = pg.sprite.Group()
        self.gold = []

        # Generate random positions for each prospector agent
        prospector_info = [utils.rand_pos("prospector") for _ in range(4)]
        self.prospectors = []
        for pos in prospector_info:
            prospector = Prospector(pos, self.space, self.all_sprites)
            self.prospectors.append(prospector)
            self.agents.append(prospector)

        banker_info = [(i + 1, utils.rand_pos("banker")) for i in range(3)]
        self.bankers = []
        for num, pos in banker_info:
            banker = Banker(pos, self.space, num, self.all_sprites)
            self.bankers.append(banker)
            self.agents.append(banker)

        # Create these dictionaries after self.agents is populated
        self.rewards = dict(zip(self.agents, [0 for _ in self.agents]))
        self.dones = dict(zip(self.agents, [False for _ in self.agents]))
        self.infos = dict(zip(self.agents, [[] for _ in self.agents]))
        self.metadata = {"render.modes": ["human"]}

        # TODO: Setup action spaces
        self.action_spaces = {}
        for a in self.agents:
            num_actions = 0
            if type(a) is Prospector:
                num_actions = 6
            else:
                num_actions = 4
            self.action_spaces[a] = spaces.Discrete(num_actions + 1)

        # TODO: Setup observation spaces
        self.observation_spaces = {}
        self.last_observation = {}
        for a in self.agents:
            self.last_observation[a] = None
            # low, high for RGB values
            self.observation_spaces[a] = spaces.Box(
                low=0, high=255, shape=const.OBSERVATION_SHAPE
            )

        """ Finish setting up environment agents """
        self.agent_order = self.agents[:]
        self._agent_selector = agent_selector(self.agent_order)
        self.agent_selection = self._agent_selector.next()

        self.banks = []
        for pos, verts in const.BANK_INFO:
            self.banks.append(Bank(pos, verts, self.space, self.all_sprites))

        for w_type, pos, verts in const.BOUNDARY_INFO:
            Fence(w_type, pos, verts, self.space, self.all_sprites)

        Water(const.WATER_INFO[0], const.WATER_INFO[1], self.space, self.all_sprites)

        # Collision Handler Functions --------------------------------------------
        # Water to Prospector
        def add_gold(arbiter, space, data):
            prospec_shape = arbiter.shapes[0]
            prospec_body = prospec_shape.body

            position = arbiter.contact_point_set.points[0].point_a
            normal = arbiter.contact_point_set.normal

            prospec_body.position = position - (24 * normal)
            prospec_body.velocity = (0, 0)

            prospec_sprite = None
            for p in self.prospectors:
                if p.body is prospec_body:
                    prospec_sprite = p
            self.rewards[prospec_sprite] += ind_reward * prospec_find_gold_reward

            for a in self.agents:
                if isinstance(a, Prospector) and a is not prospec_sprite:
                    self.rewards[a] += group_reward * prospec_find_gold_reward
                elif isinstance(a, Banker):
                    self.rewards[a] += other_group_reward * prospec_find_gold_reward

            if prospec_body.nugget is None:
                position = arbiter.contact_point_set.points[0].point_a

                gold = Gold(position, prospec_body, self.space)
                self.gold.append(gold)
                prospec_body.nugget = gold

            return True

        # Prospector to banker
        def handoff_gold_handler(arbiter, space, data):
            banker_shape, gold_shape = arbiter.shapes[0], arbiter.shapes[1]

            gold_sprite = None
            for g in self.gold:
                if g.id == gold_shape.id:
                    gold_sprite = g

            if gold_sprite.parent_body.sprite_type == "prospector":
                banker_body = banker_shape.body
                prospec_body = gold_sprite.parent_body

                prospec_sprite = None
                for p in self.prospectors:
                    if p.body is prospec_body:
                        prospec_sprite = p
                self.rewards[prospec_sprite] += prospec_handoff_gold_reward

                for a in self.agents:
                    if isinstance(a, Prospector) and a is not prospec_sprite:
                        self.rewards[a] += group_reward * prospec_handoff_gold_reward
                    elif isinstance(a, Banker):
                        self.rewards[a] += (
                            other_group_reward * prospec_handoff_gold_reward
                        )

                banker_sprite = None
                for b in self.bankers:
                    if b.shape is banker_shape:
                        banker_sprite = b
                self.rewards[banker_sprite] += banker_receive_gold_reward

                for a in self.agents:
                    if isinstance(a, Prospector):
                        self.rewards[a] += (
                            other_group_reward * banker_receive_gold_reward
                        )
                    elif isinstance(a, Banker) and a is not banker_sprite:
                        self.rewards[a] += group_reward * banker_receive_gold_reward

                normal = arbiter.contact_point_set.normal
                # Correct the angle because banker's head is rotated pi/2
                corrected = utils.normalize_angle(banker_body.angle + (math.pi / 2))
                if (
                    corrected - const.BANKER_HANDOFF_TOLERANCE
                    <= normal.angle
                    <= corrected + const.BANKER_HANDOFF_TOLERANCE
                ):
                    gold_sprite.parent_body.nugget = None

                    gold_sprite.parent_body = banker_body
                    banker_body.nugget = gold_sprite
                    banker_body.nugget_offset = normal.angle

            return True

        # Banker to chest
        def gold_score_handler(arbiter, space, data):
            gold_shape, chest = arbiter.shapes[0], arbiter.shapes[1]

            gold_class = None
            for g in self.gold:
                if g.id == gold_shape.id:
                    gold_class = g

            if gold_class.parent_body.sprite_type == "banker":
                chest.body.score += 1
                self.space.remove(gold_shape, gold_shape.body)
                gold_class.parent_body.nugget = None
                banker_body = gold_class.parent_body

                banker_sprite = None
                for b in self.bankers:
                    if b.body is banker_body:
                        banker_sprite = b
                self.rewards[banker_sprite] += banker_deposit_gold_reward

                for a in self.agents:
                    if isinstance(a, Prospector):
                        self.rewards[a] += (
                            other_group_reward * banker_deposit_gold_reward
                        )
                    elif isinstance(a, Banker) and a is not banker_sprite:
                        self.rewards[a] += group_reward * banker_deposit_gold_reward

                # total_score = ", ".join(
                #     [
                #         "Chest %d: %d" % (i, c.body.score)
                #         for i, c in enumerate(self.chests)
                #     ]
                # )

                # print(total_score)
                self.gold.remove(gold_class)
                self.all_sprites.remove(gold_class)

            return False

        # Create the collision event generators
        gold_dispenser = self.space.add_collision_handler(
            CollisionTypes.PROSPECTOR, CollisionTypes.WATER
        )

        gold_dispenser.begin = add_gold

        handoff_gold = self.space.add_collision_handler(
            CollisionTypes.BANKER, CollisionTypes.GOLD
        )

        handoff_gold.begin = handoff_gold_handler

        gold_score = self.space.add_collision_handler(
            CollisionTypes.GOLD, CollisionTypes.CHEST
        )

        gold_score.begin = gold_score_handler

    def observe(self, agent):
        capture = pg.surfarray.array3d(self.screen)
        a = self.agents[agent]

        delta = const.OBSERVATION_SIDE_LENGTH // 2
        center = a.center  # Calculated property added to prospector and banker classes
        x, y = center
        sub_screen = capture[x - delta: x + delta, y - delta: y + delta, :]

        return sub_screen

    def step(self, action, observe=True):
        agent = self.agent_selection
        # TODO: Figure out rewards
        if action is None:
            print("Error: NoneType received as action")
        else:
            agent.update(action)

        all_agents_updated = self._agent_selector.is_last()
        # Only take next step in game if all agents have received an action
        if all_agents_updated:
            self.frame += 1
            # If we reached max frames, we're done
            if self.frame == self.max_frames:
                self.dones = dict(zip(self.agents, [True for _ in self.agents]))

            if self.rendering:
                self.clock.tick(const.FPS)
            else:
                self.clock.tick()
            self.space.step(1 / 15)

        self.agent_selection = self._agent_selector.next()
        observation = self.observe(self.agent_selection)

        pg.event.pump()
        if observe:
            return observation

    def reward(self):
        return self.rewards

    def reset(self, observe=True):
        self.done = False
        self.agents = []

        # Re-create all agents and Pymunk space
        self.space = pm.Space()
        self.space.gravity = Vec2d(0.0, 0.0)
        self.space.damping = 0.0

        self.all_sprites = pg.sprite.Group()
        self.gold = []

        prospector_info = [utils.rand_pos("prospector") for _ in range(4)]
        self.prospectors = []
        for pos in prospector_info:
            prospector = Prospector(pos, self.space, self.all_sprites)
            self.prospectors.append(prospector)
            self.agents.append(prospector)

        banker_info = [(i + 1, utils.rand_pos("banker")) for i in range(3)]
        self.bankers = []
        for num, pos in banker_info:
            banker = Banker(pos, self.space, num, self.all_sprites)
            self.bankers.append(banker)
            self.agents.append(banker)

        self.banks = []
        for pos, verts in const.BANK_INFO:
            self.banks.append(Bank(pos, verts, self.space, self.all_sprites))

        for w_type, pos, verts in const.BOUNDARY_INFO:
            Fence(w_type, pos, verts, self.space, self.all_sprites)

        Water(const.WATER_INFO[0], const.WATER_INFO[1], self.space, self.all_sprites)

        self.rewards = dict(zip(self.agents, [0 for _ in self.agents]))
        self.dones = dict(zip(self.agents, [False for _ in self.agents]))
        self.infos = dict(zip(self.agents, [[] for _ in self.agents]))
        self.metadata = {"render.modes": ["human"]}

        self.rendering = False
        self.frame = 0
        self._agent_selector.reinit(self.agent_order)
        self.agent_selection = self._agent_selector.next()

        if observe:
            return self.observe(self.agent_selection)

    def render(self, mode="human"):
        if not self.rendering:
            pg.display.init()
            self.screen = pg.display.set_mode(const.SCREEN_SIZE)
            self.rendering = True
        self.draw()
        pg.display.flip()

    def draw(self):
        self.screen.blit(self.background, self.background_rect)
        self.all_sprites.draw(self.screen)

        for p in self.prospectors:
            if p.body.nugget is not None:
                p.body.nugget.draw(self.screen)

        for b in self.bankers:
            if b.body.nugget is not None:
                b.body.nugget.draw(self.screen)

    def close(self):
        pg.event.pump()
        pg.display.quit()
        pg.quit()


if __name__ == "__main__":
    pg.init()
    game = Game()
    game.run()



import random


def test_render(env):
    render_modes = env.metadata.get('render.modes')
    assert render_modes is not None, "Environment's that support rendering must define render modes in metadata"
    env.reset(observe=False)
    for mode in render_modes:
        for _ in range(10):
            for agent in env.agent_order:
                if 'legal_moves' in env.infos[agent]:
                    action = random.choice(env.infos[agent]['legal_moves'])
                else:
                    action = env.action_spaces[agent].sample()
                env.step(action, observe=False)
                env.render(mode=mode)
                if all(env.dones.values()):
                    env.reset()
                    break



import math

SCREEN_SIZE = SCREEN_WIDTH, SCREEN_HEIGHT = (1280, 720)
BACKGROUND_COLOR = (217, 151, 106)
FPS = 15

# FENCE_WIDTH = 100
FENCE_WIDTH = 22
WATER_HEIGHT = 100

# AGENT_RADIUS = 16.5
AGENT_RADIUS = 15
AGENT_DIAMETER = AGENT_RADIUS * 2
# The 3 is for RBG values
OBSERVATION_SIDE_LENGTH = 5 * AGENT_DIAMETER
OBSERVATION_SHAPE = (OBSERVATION_SIDE_LENGTH, OBSERVATION_SIDE_LENGTH, 3)


MAX_SPRITE_ROTATION = math.pi / 4

PROSPECTOR_SPEED = 150
BANKER_SPEED = 100
BANKER_HANDOFF_TOLERANCE = math.pi / 4
TWO_PI = math.pi * 2.0

FENCE_COLLISION_BUFFER = AGENT_DIAMETER
VERT_FENCE_HEIGHT = SCREEN_HEIGHT - WATER_HEIGHT
# VERT_FENCE_WIDTH =

# For the left and right fences
FENCE_VERT_VERTICES = (
    (0, 0),
    (FENCE_WIDTH + FENCE_COLLISION_BUFFER, 0),
    (FENCE_WIDTH + FENCE_COLLISION_BUFFER, VERT_FENCE_HEIGHT),
    (0, VERT_FENCE_HEIGHT),
)

# For the top fence
FENCE_HORIZ_VERTICES = (
    (0, 0),
    (SCREEN_WIDTH, 0),
    (SCREEN_WIDTH, FENCE_WIDTH + FENCE_COLLISION_BUFFER),
    (0, FENCE_WIDTH + FENCE_COLLISION_BUFFER),
)

FENCE_INFO = [
    ("left", [0, 0], [0, 0],FENCE_VERT_VERTICES),  # left boundary
    ("top", [0, 0], [0, 0], FENCE_HORIZ_VERTICES),  # top boundary
    ("right", [SCREEN_WIDTH - FENCE_WIDTH, 0], [SCREEN_WIDTH - (FENCE_WIDTH + FENCE_COLLISION_BUFFER), 0], FENCE_VERT_VERTICES),
]

BANK_SIZE = BANK_WIDTH, BANK_HEIGHT = 184, 100

BANK_VERTS = (
    (0, 0),
    (BANK_WIDTH, 0),
    (BANK_WIDTH, BANK_HEIGHT),
    (0, BANK_HEIGHT),
)

BANK_INFO = [
    ([184 * 1, 50], BANK_VERTS),
    ([184 * 3, 50], BANK_VERTS),
    ([184 * 5, 50], BANK_VERTS),
]

WATER_INFO = [
    (0, SCREEN_HEIGHT - WATER_HEIGHT), # position
    (   # vertices
        (0, 0),
        (SCREEN_WIDTH, 0),
        (SCREEN_WIDTH, WATER_HEIGHT),
        (0, WATER_HEIGHT),
    ),
]


from pettingzoo import AECEnv
from gym import spaces
import numpy as np
import warnings

from .manual_control import manual_control


class env(AECEnv):
    metadata = {'render.modes': ['ansi']}

    def __init__(self):
        super(env, self).__init__()
        # 6 rows x 7 columns
        # blank space = 0
        # agent 0 -- 1
        # agent 1 -- 2
        # flat representation in row major order
        self.board = [0] * (6 * 7)

        self.num_agents = 2
        self.agents = ['player_0', 'player_1']

        self.agent_order = list(self.agents)

        self.action_spaces = {i: spaces.Discrete(7) for i in range(2)}
        self.observation_spaces = {i: spaces.Box(low=0, high=2, shape=(6, 7), dtype=np.int8) for i in range(2)}

        self.rewards = {i: 0 for i in range(self.num_agents)}
        self.dones = {i: False for i in range(self.num_agents)}
        self.infos = {i: {'legal_moves': list(range(7))} for i in range(self.num_agents)}

        self.agent_selection = 0

    # Key
    # ----
    # blank space = 0
    # agent 0 = 1
    # agent 1 = 2
    # An observation is list of lists, where each list represents a row
    #
    # array([[0, 1, 1, 2, 0, 1, 0],
    #        [1, 0, 1, 2, 2, 2, 1],
    #        [0, 1, 0, 0, 1, 2, 1],
    #        [1, 0, 2, 0, 1, 1, 0],
    #        [2, 0, 0, 0, 1, 1, 0],
    #        [1, 1, 2, 1, 0, 1, 0]], dtype=int8)
    def observe(self, agent):
        return np.array(self.board).reshape(6, 7)

    # action in this case is a value from 0 to 6 indicating position to move on the flat representation of the connect4 board
    def step(self, action, observe=True):
        # check if input action is a valid move (0 == empty spot)
        if(self.board[0:7][action] == 0):
            # valid move
            for i in list(filter(lambda x: x % 7 == action, list(range(41, -1, -1)))):
                if self.board[i] == 0:
                    self.board[i] = self.agent_selection + 1
                    break

            next_agent = 1 if (self.agent_selection == 0) else 0

            # update infos with valid moves
            self.infos[self.agent_selection]['legal_moves'] = [i for i in range(7) if self.board[i] == 0]
            self.infos[next_agent]['legal_moves'] = [i for i in range(7) if self.board[i] == 0]

            winner = self.check_for_winner()

            # check if there is a winner
            if winner:
                self.rewards[self.agent_selection] += 1
                self.rewards[next_agent] -= 1
                self.dones = {i: True for i in range(self.num_agents)}
            # check if there is a tie
            elif all(x in [1, 2] for x in self.board):
                # once either play wins or there is a draw, game over, both players are done
                self.dones = {i: True for i in range(self.num_agents)}
            else:
                # no winner yet
                self.agent_selection = next_agent

        else:
            # invalid move, end game
            self.rewards[self.agent_selection] -= 1
            self.dones = {i: True for i in range(self.num_agents)}
            warnings.warn("Bad connect four move made, game terminating with current player losing. env.infos[player]['legal_moves'] contains a list of all legal moves that can be chosen.")

        if observe:
            return self.observe(self.agent_selection)
        else:
            return

    def reset(self, observe=True):
        # reset environment
        self.board = [0] * (6 * 7)

        self.rewards = {i: 0 for i in range(self.num_agents)}
        self.dones = {i: False for i in range(self.num_agents)}
        self.infos = {i: {'legal_moves': list(range(7))} for i in range(self.num_agents)}

        # selects the first agent
        self.agent_selection = 0
        if observe:
            return self.observe(self.agent_selection)
        else:
            return

    def render(self, mode='ansi'):
        print(str(self.observe(self.agent_selection)))

    def close(self):
        pass

    def check_for_winner(self):
        board = np.array(self.board).reshape(6, 7)
        piece = self.agent_selection + 1

        # Check horizontal locations for win
        column_count = 7
        row_count = 6

        for c in range(column_count - 3):
            for r in range(row_count):
                if board[r][c] == piece and board[r][c + 1] == piece and board[r][c + 2] == piece and board[r][c + 3] == piece:
                    return True

        # Check vertical locations for win
        for c in range(column_count):
            for r in range(row_count - 3):
                if board[r][c] == piece and board[r + 1][c] == piece and board[r + 2][c] == piece and board[r + 3][c] == piece:
                    return True

        # Check positively sloped diaganols
        for c in range(column_count - 3):
            for r in range(row_count - 3):
                if board[r][c] == piece and board[r + 1][c + 1] == piece and board[r + 2][c + 2] == piece and board[r + 3][c + 3] == piece:
                    return True

        # Check negatively sloped diaganols
        for c in range(column_count - 3):
            for r in range(3, row_count):
                if board[r][c] == piece and board[r - 1][c + 1] == piece and board[r - 2][c + 2] == piece and board[r - 3][c + 3] == piece:
                    return True

        return False



import gym
import time
from itertools import count
import random
import numpy as np
from gym_backgammon.envs.backgammon import WHITE, BLACK, COLORS, TOKEN

env = gym.make('gym_backgammon:backgammon-v0')
# env = gym.make('gym_backgammon:backgammon-pixel-v0')

random.seed(0)
np.random.seed(0)


class RandomAgent:
    def __init__(self, color):
        self.color = color
        self.name = 'AgentExample({})'.format(self.color)

    def roll_dice(self):
        return (-random.randint(1, 6), -random.randint(1, 6)) if self.color == WHITE else (random.randint(1, 6), random.randint(1, 6))

    def choose_best_action(self, actions, env):
        return random.choice(list(actions)) if actions else None


def make_plays():
    wins = {WHITE: 0, BLACK: 0}

    agents = {WHITE: RandomAgent(WHITE), BLACK: RandomAgent(BLACK)}

    agent_color, first_roll, observation = env.reset()

    agent = agents[agent_color]

    t = time.time()

    env.render(mode='human')

    for i in count():
        if first_roll:
            roll = first_roll
            first_roll = None
        else:
            roll = agent.roll_dice()

        print("Current player={} ({} - {}) | Roll={}".format(agent.color, TOKEN[agent.color], COLORS[agent.color], roll))

        actions = env.get_valid_actions(roll)

        # print("\nLegal Actions:")
        # for a in actions:
        #     print(a)

        action = agent.choose_best_action(actions, env)

        observation_next, reward, done, winner = env.step(action)

        env.render(mode='human')

        if done:
            if winner is not None:
                wins[winner] += 1

            tot = wins[WHITE] + wins[BLACK]
            tot = tot if tot > 0 else 1

            print("Game={} | Winner={} after {:<4} plays || Wins: {}={:<6}({:<5.1f}%) | {}={:<6}({:<5.1f}%) | Duration={:<.3f} sec".format(1, winner, i,
                agents[WHITE].name, wins[WHITE], (wins[WHITE] / tot) * 100,
                agents[BLACK].name, wins[BLACK], (wins[BLACK] / tot) * 100, time.time() - t))

            break

        agent_color = env.get_opponent_agent()
        agent = agents[agent_color]
        observation = observation_next

    env.close()


if __name__ == '__main__':
    make_plays()



def manual_control(**kwargs):
    from .connect_four import env as _env
    env = _env(**kwargs)

    done = all(env.dones.values())

    print("valid actions are column indices from 0 - 6")
    while not done:
        current_player = env.agent_selection + 1

        action = int(input(f"Player {current_player}: "))

        # check if its a valid action
        if env.board[action] == 0:
            obs = env.step(action)
            print(obs)

            done = all(env.dones.values())

            if done:
                print(f"Player {current_player} won!")
                break
        else:
            print("Invalid move")



import os


# add more dir names here
dir_names = ["gamma", "sisl", "magent", "mpe"]

for name in dir_names:
    root_dir = os.path.join(os.pardir, name)

    for _dir, subdirs, files in os.walk(root_dir):
        for file in files:
            if file.endswith(".py"):
                with open(os.path.join(_dir, file), 'r') as f:
                    for line in f:
                        if line.lstrip().startswith("print"):
                            print("File: {} has a print statement. Please remove it.".format(os.path.join(_dir, file)))
                            break



from pettingzoo import AECEnv
from pettingzoo.utils import agent_selector
from gym import spaces
import numpy as np
import warnings

from .manual_control import manual_control

from .board import Board


class env(AECEnv):
    metadata = {'render.modes': ['human']}

    def __init__(self):
        super(env, self).__init__()
        self.board = Board()

        self.num_agents = 2
        self.agents = ["player_1", "player_2"]

        self.agent_order = list(self.agents)

        self.action_spaces = {i: spaces.Discrete(9) for i in self.agents}
        self.observation_spaces = {i: spaces.Box(low=0, high=2, shape=(3, 3), dtype=np.int8) for i in self.agents}

        self.rewards = {i: 0 for i in self.agents}
        self.dones = {i: False for i in self.agents}
        self.infos = {i: {'legal_moves': list(range(0, 9))} for i in self.agents}

        self._agent_selector = agent_selector(self.agent_order)
        self.agent_selection = self._agent_selector.reset()

    # Key
    # ----
    # blank space = 0
    # agent 0 = 1
    # agent 1 = 2
    # An observation is list of lists, where each list represents a row
    #
    # [[0,0,2]
    #  [1,2,1]
    #  [2,1,0]]
    def observe(self, agent):
        # return observation of an agent
        s = np.array(self.board.squares)
        return s.reshape(3, 3).T

    # action in this case is a value from 0 to 8 indicating position to move on tictactoe board
    def step(self, action, observe=True):
        # check if input action is a valid move (0 == empty spot)
        if(self.board.squares[action] == 0):
            # play turn
            self.board.play_turn(self.agents.index(self.agent_selection), action)

            # update infos
            # list of valid actions (indexes in board)
            # next_agent = self.agents[(self.agents.index(self.agent_selection) + 1) % len(self.agents)]
            next_agent = self._agent_selector.next()
            self.infos[self.agent_selection]['legal_moves'] = [i for i in range(len(self.board.squares)) if self.board.squares[i] == 0]
            self.infos[next_agent]['legal_moves'] = [i for i in range(len(self.board.squares)) if self.board.squares[i] == 0]

            if self.board.check_game_over():
                winner = self.board.check_for_winner()

                if winner == -1:
                    # tie
                    pass
                elif winner == 1:
                    # agent 0 won
                    self.rewards[self.agents[0]] += 1
                    self.rewards[self.agents[1]] -= 1
                else:
                    # agent 1 won
                    self.rewards[self.agents[1]] += 1
                    self.rewards[self.agents[0]] -= 1

                # once either play wins or there is a draw, game over, both players are done
                self.dones = {i: True for i in self.agents}

        else:
            # invalid move, end game
            self.rewards[self.agent_selection] -= 1
            self.dones = {i: True for i in self.agents}
            warnings.warn("Bad tictactoe move made, game terminating with current player losing. env.infos[player]['legal_moves'] contains a list of all legal moves that can be chosen.")

        # Switch selection to next agents
        self.agent_selection = next_agent

        if observe:
            return self.observe(self.agent_selection)
        else:
            return

    def reset(self, observe=True):
        # reset environment
        self.board = Board()

        self.rewards = {i: 0 for i in self.agents}
        self.dones = {i: False for i in self.agents}
        self.infos = {i: {'legal_moves': list(range(0, 9))} for i in self.agents}

        # selects the first agent
        self._agent_selector.reset()
        self.agent_selection = self._agent_selector.reset()
        if observe:
            return self.observe(self.agent_selection)
        else:
            return

    def render(self, mode='human'):
        def getSymbol(input):
            if input == 0:
                return '-'
            elif input == 1:
                return 'X'
            else:
                return 'O'

        board = list(map(getSymbol, self.board.squares))

        print(" " * 5 + "|" + " " * 5 + "|" + " " * 5)
        print(f"  {board[0]}  " + "|" + f"  {board[3]}  " + "|" + f"  {board[6]}  ")
        print("_" * 5 + "|" + "_" * 5 + "|" + "_" * 5)

        print(" " * 5 + "|" + " " * 5 + "|" + " " * 5)
        print(f"  {board[1]}  " + "|" + f"  {board[4]}  " + "|" + f"  {board[7]}  ")
        print("_" * 5 + "|" + "_" * 5 + "|" + "_" * 5)

        print(" " * 5 + "|" + " " * 5 + "|" + " " * 5)
        print(f"  {board[2]}  " + "|" + f"  {board[5]}  " + "|" + f"  {board[8]}  ")
        print(" " * 5 + "|" + " " * 5 + "|" + " " * 5)

    def close(self):
        pass



from pettingzoo import AECEnv
from pettingzoo.utils.agent_selector import agent_selector
from pettingzoo.utils.env_logger import EnvLogger
from gym import spaces
import rlcard
import random
from rlcard.utils.utils import print_card
import numpy as np


class env(AECEnv):

    metadata = {'render.modes': ['human']}

    def __init__(self, seed=None, **kwargs):
        super(env, self).__init__()
        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)
        self.env = rlcard.make('limit-holdem', **kwargs)
        self.agents = ['player_0', 'player_1']
        self.num_agents = len(self.agents)
        self.has_reset = False

        self.observation_spaces = self._convert_to_dict([spaces.Box(low=0.0, high=1.0, shape=(72,), dtype=np.bool) for _ in range(self.num_agents)])
        self.action_spaces = self._convert_to_dict([spaces.Discrete(self.env.game.get_action_num()) for _ in range(self.num_agents)])

        self.agent_order = self.agents
        self._agent_selector = agent_selector(self.agent_order)

    def _int_to_name(self, ind):
        return self.agents[ind]

    def _name_to_int(self, name):
        return self.agents.index(name)

    def _convert_to_dict(self, list_of_list):
        return dict(zip(self.agents, list_of_list))

    def observe(self, agent):
        if not self.has_reset:
            EnvLogger.error_observe_before_reset()
        obs = self.env.get_state(self._name_to_int(agent))
        return obs['obs']

    def step(self, action, observe=True):
        if not self.has_reset:
            EnvLogger.error_step_before_reset()
        backup_policy = "Game terminating with current player losing"
        act_space = self.action_spaces[self.agent_selection]
        if np.isnan(action).any():
            EnvLogger.warn_action_is_NaN(backup_policy)
        if not act_space.contains(action):
            EnvLogger.warn_action_out_of_bound(action, act_space, backup_policy)

        if self.dones[self.agent_selection]:
            self.dones = self._convert_to_dict([True for _ in range(self.num_agents)])
            obs = False
        else:
            if action not in self.infos[self.agent_selection]['legal_moves']:
                EnvLogger.warn_on_illegal_move()
                self.rewards[self.agent_selection] = -1
                self.dones = self._convert_to_dict([True for _ in range(self.num_agents)])
                info_copy = self.infos[self.agent_selection]
                self.infos = self._convert_to_dict([{'legal_moves': [2]} for agent in range(self.num_agents)])
                self.infos[self.agent_selection] = info_copy
                self.agent_selection = self._agent_selector.next()
                return self._last_obs
            obs, next_player_id = self.env.step(action)
            self._last_obs = obs['obs']
            if self.env.is_over():
                self.dones = self._convert_to_dict([True for _ in range(self.num_agents)])
                self.rewards = self._convert_to_dict(self.env.get_payoffs())
                self.infos[self._int_to_name(next_player_id)]['legal_moves'] = [2]
            else:
                self.infos[self._int_to_name(next_player_id)]['legal_moves'] = obs['legal_actions']
        self.agent_selection = self._agent_selector.next()
        if observe:
            return obs['obs'] if obs else self._last_obs

    def reset(self, observe=True):
        self.has_reset = True
        obs, player_id = self.env.init_game()
        self.agent_order = [self._int_to_name(agent) for agent in [player_id, 0 if player_id == 1 else 1]]
        self._agent_selector.reinit(self.agent_order)
        self.agent_selection = self._agent_selector.reset()
        self.rewards = self._convert_to_dict(np.array([0.0, 0.0]))
        self.dones = self._convert_to_dict([False for _ in range(self.num_agents)])
        self.infos = self._convert_to_dict([{'legal_moves': []} for _ in range(self.num_agents)])
        self.infos[self._int_to_name(player_id)]['legal_moves'] = obs['legal_actions']
        self._last_obs = obs['obs']
        if observe:
            return obs['obs']
        else:
            return

    def render(self, mode='human'):
        for player in self.agents:
            state = self.env.game.get_state(self._name_to_int(player))
            print("\n=============== {}'s Hand ===============".format(player))
            print_card(state['hand'])
            print("\n{}'s Chips: {}".format(player, state['my_chips']))
        print('\n================= Public Cards =================')
        print_card(state['public_cards']) if state['public_cards'] else print('No public cards.')
        print('\n')

    def close(self):
        EnvLogger.warn_close_unrendered_env()



from .pursuit_base import Pursuit as _env
from .manual_control import manual_control
from pettingzoo import AECEnv
from pettingzoo.utils import agent_selector
import numpy as np
from pettingzoo.utils import EnvLogger
import pygame


class env(AECEnv):

    metadata = {'render.modes': ['human']}

    def __init__(self, seed=None, *args, **kwargs):
        super(env, self).__init__()
        self.env = _env(*args, seed, **kwargs)
        pygame.init()
        self.num_agents = self.env.num_agents
        self.agents = ["pursuer_" + str(a) for a in range(self.num_agents)]
        self.agent_name_mapping = dict(zip(self.agents, list(range(self.num_agents))))
        self.agent_order = self.agents[:]
        self._agent_selector = agent_selector(self.agent_order)
        self.has_reset = False
        # spaces
        self.n_act_agents = self.env.act_dims[0]
        self.action_spaces = dict(zip(self.agents, self.env.action_space))
        self.observation_spaces = dict(
            zip(self.agents, self.env.observation_space))
        self.steps = 0
        self.display_wait = 0.0
        self.closed = False

    def reset(self, observe=True):
        self.has_reset = True
        self.steps = 0
        self.rewards = dict(
            zip(self.agents, [np.float64(0) for _ in self.agents]))
        self.dones = dict(zip(self.agents, [False for _ in self.agents]))
        self.infos = dict(zip(self.agents, [{} for _ in self.agents]))
        self._agent_selector.reinit(self.agent_order)
        self.agent_selection = self._agent_selector.next()
        self.env.reset()
        if observe:
            return self.observe(self.agent_selection)

    def close(self):
        if not self.has_reset:
            EnvLogger.warn_close_before_reset()
        elif not self.closed and self.has_reset:
            self.closed = True
            self.env.close()

    def render(self, mode="human"):
        if not self.has_reset:
            EnvLogger.error_render_before_reset()
        elif not self.closed:
            self.env.render()

    def step(self, action, observe=True):
        if not self.has_reset:
            EnvLogger.error_step_before_reset()
        agent = self.agent_selection
        if action is None or np.isnan(action):
            action = 0
            EnvLogger.warn_action_is_NaN(backup_policy="setting action to 0")
        elif not self.action_spaces[agent].contains(action):
            EnvLogger.warn_action_out_of_bound(action=action, action_space=self.action_spaces[agent], backup_policy="setting action to 0")
            action = 0
        self.env.step(action, self.agent_name_mapping[agent], self._agent_selector.is_last())
        for k in self.dones:
            if self.env.frames >= self.env.max_frames:
                self.dones[k] = True
            else:
                self.dones[k] = self.env.is_terminal
        for k in self.agents:
            self.rewards[k] = self.env.latest_reward_state[self.agent_name_mapping[k]]
        self.steps += 1
        self.agent_selection = self._agent_selector.next()
        if observe:
            return self.observe(self.agent_selection)

    def observe(self, agent):
        if not self.has_reset:
            EnvLogger.error_observe_before_reset()
        o = np.array(self.env.safely_observe(self.agent_name_mapping[agent]))
        return o



from .knights_archers_zombies import knights_archers_zombies as knights_archers_zombies_v0
from .pistonball import pistonball as pistonball_v0
from .cooperative_pong import cooperative_pong as cooperative_pong_v0
from .prison import prison as prison_v0



import pyglet
from pyglet import gl
import numpy as np
from collections import namedtuple
from gym_backgammon.envs.backgammon import WHITE, BLACK, NUM_POINTS
import os
import platform
Coords = namedtuple('Coords', ['x', 'y'])

SCALING = 2 if platform.system() == 'Darwin' else 1

class Viewer(object):
    def __init__(self, width, height):
        self.width = width
        self.height = height

        self.window = pyglet.window.Window(width=width, height=height, display=None)

        pyglet.resource.path = [os.path.dirname(__file__) + '/resources']
        pyglet.resource.reindex()

        empty_board_image = pyglet.resource.image("board.png")
        empty_board_image.width = width
        empty_board_image.height = height

        self.empty_board_image = empty_board_image

        # CHECKERS
        self.checker_diameter = self.width / 15  # 40

        self.checkers = {
            WHITE: {i: pyglet.resource.image("white_{}.png".format(i)) for i in range(1, 16)},
            BLACK: {i: pyglet.resource.image("black_{}.png".format(i)) for i in range(1, 16)}
        }

        coords = {}
        shifts = [2, 3, 4, 5, 6, 7, 9, 10, 11, 12, 13, 14]

        for i in range(NUM_POINTS):
            index = 23 - i if 11 < i <= 23 else i

            if i < 12:
                coords[i] = Coords(x=width - (shifts[index] * self.checker_diameter), y=self.checker_diameter)
            else:
                coords[i] = Coords(x=width - (shifts[index] * self.checker_diameter), y=height - self.checker_diameter)

        coords['bar_{}'.format(WHITE)] = Coords(x=width - (8 * self.checker_diameter), y=height - self.checker_diameter)
        coords['bar_{}'.format(BLACK)] = Coords(x=width - (8 * self.checker_diameter), y=self.checker_diameter)

        coords['off_{}'.format(WHITE)] = Coords(x=width - self.checker_diameter, y=self.checker_diameter)
        coords['off_{}'.format(BLACK)] = Coords(x=width - self.checker_diameter, y=height - self.checker_diameter)

        self.points_coord = coords

        # remove
        self.counter = 0

        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

    def close(self):
        self.window.close()

    def render(self, board, bar, off, state_w, state_h, return_rgb_array=True):
        gl.glClearColor(1, 1, 1, 1)

        self.window.switch_to()
        self.window.dispatch_events()
        self.window.clear()

        batch = pyglet.graphics.Batch()
        sprites = []

        for point, (checkers, player) in enumerate(board):

            if player is not None:
                assert player in [WHITE, BLACK], print("Should be WHITE (0) or BLACK (1), not {}".format(player))
                assert point in self.points_coord, print("Should be 0 <= point < 24, not {}".format(point))
                assert 0 <= checkers <= 15, print("Should be 0 <= checkers <= 15, not {}".format(checkers))

                c = self.points_coord[point]
                img = self.checkers[player][checkers]
                checkers = 5 if checkers > 5 else checkers
                img.width = self.checker_diameter
                img.height = self.checker_diameter * checkers

                if point < 12:
                    s = pyglet.sprite.Sprite(img=img, x=c.x, y=c.y, batch=batch)
                else:
                    s = pyglet.sprite.Sprite(img=img, x=c.x + self.checker_diameter, y=c.y - (checkers * self.checker_diameter) + img.height, batch=batch)
                    s.rotation = 180

                sprites.append(s)

        for player in [WHITE, BLACK]:
            # BAR
            checkers = bar[player]
            if checkers > 0:
                c = self.points_coord['bar_{}'.format(player)]

                img = self.checkers[player][checkers]
                checkers = 5 if checkers > 5 else checkers
                img.width = self.checker_diameter
                img.height = self.checker_diameter * checkers

                if player == BLACK:
                    s = sprites.append(pyglet.sprite.Sprite(img=img, x=c.x, y=c.y, batch=batch))
                else:
                    s = pyglet.sprite.Sprite(img=img, x=c.x + self.checker_diameter, y=c.y - (checkers * self.checker_diameter) + img.height, batch=batch)
                    s.rotation = 180
                sprites.append(s)

            # OFF
            checkers = off[player]
            if checkers > 0:

                c = self.points_coord['off_{}'.format(player)]
                img = self.checkers[player][checkers]
                checkers = 5 if checkers > 5 else checkers
                img.width = self.checker_diameter
                img.height = self.checker_diameter * checkers

                if player == WHITE:
                    s = sprites.append(pyglet.sprite.Sprite(img=img, x=c.x, y=c.y, batch=batch))
                else:
                    s = pyglet.sprite.Sprite(img=img, x=c.x + self.checker_diameter, y=c.y - (checkers * self.checker_diameter) + img.height, batch=batch)
                    s.rotation = 180

                sprites.append(s)

        gl.glViewport(0, 0, state_w, state_h)

        pyglet.sprite.Sprite(img=self.empty_board_image, batch=None).draw()
        batch.draw()

        arr = None

        if return_rgb_array:
            buffer = pyglet.image.get_buffer_manager().get_color_buffer()
            image_data = buffer.get_image_data()
            arr = np.fromstring(image_data.data, dtype=np.uint8, sep='')
            arr = arr.reshape((state_h, state_w, 4))
            arr = arr[::-1, :, 0:3]

        gl.glViewport(0, 0, SCALING * self.window.width, SCALING * self.window.height)
        pyglet.sprite.Sprite(img=self.empty_board_image, batch=None).draw()
        batch.draw()

        self.window.flip()

        return arr



from . import chess_utils
import chess
from pettingzoo import AECEnv
from gym import spaces
import numpy as np
import warnings
from pettingzoo.utils.agent_selector import agent_selector
from pettingzoo.utils.env_logger import EnvLogger


class env(AECEnv):

    metadata = {'render.modes': ['human', 'ascii']}

    def __init__(self):
        super(env, self).__init__()

        self.board = chess.Board()

        self.agents = ["player_{}".format(i) for i in range(2)]

        self.agent_order = list(self.agents)
        self._agent_selector = agent_selector(self.agent_order)

        self.action_spaces = {name: spaces.Discrete(8 * 8 * 73) for name in self.agents}
        self.observation_spaces = {name: spaces.Box(low=0, high=1, shape=(8, 8, 20), dtype=np.float32) for name in self.agents}

        # self.rewards = None
        # self.dones = None
        # self.infos = None
        #
        # self.agent_selection = None

        self.has_reset = False
        self.has_rendered = False

        self.num_agents = len(self.agents)

    def observe(self, agent):
        if not self.has_reset:
            EnvLogger.error_observe_before_reset()
        return chess_utils.get_observation(self.board, self.agents.index(agent))

    def reset(self, observe=True):
        self.has_reset = True

        self.board = chess.Board()

        self.agent_selection = self._agent_selector.reset()

        self.rewards = {name: 0 for name in self.agents}
        self.dones = {name: False for name in self.agents}
        self.infos = {name: {'legal_moves': []} for name in self.agents}
        self.infos[self.agent_selection]['legal_moves'] = chess_utils.legal_moves(self.board)

        if observe:
            return self.observe(self.agent_selection)
        else:
            return

    def set_game_result(self, result_val):
        for i, name in enumerate(self.agents):
            self.dones[name] = True
            result_coef = 1 if i == 0 else -1
            self.rewards[name] = result_val * result_coef
            self.infos[name] = {'legal_moves': []}

    def step(self, action, observe=True):
        if not self.has_reset:
            EnvLogger.error_step_before_reset()
        backup_policy = "game terminating with current player losing"
        act_space = self.action_spaces[self.agent_selection]
        if np.isnan(action).any():
            EnvLogger.warn_action_is_NaN(backup_policy)
        if not act_space.contains(action):
            EnvLogger.warn_action_out_of_bound(action, act_space, backup_policy)

        current_agent = self.agent_selection
        current_index = self.agents.index(current_agent)
        self.agent_selection = next_agent = self._agent_selector.next()

        old_legal_moves = self.infos[current_agent]['legal_moves']

        if action not in old_legal_moves:
            EnvLogger.warn_on_illegal_move()
            player_loses_val = -1 if current_index == 0 else 1
            self.set_game_result(player_loses_val)
            self.rewards[next_agent] = 0
        else:
            chosen_move = chess_utils.action_to_move(self.board, action, current_index)
            assert chosen_move in self.board.legal_moves
            self.board.push(chosen_move)

            next_legal_moves = chess_utils.legal_moves(self.board)

            is_stale_or_checkmate = not any(next_legal_moves)

            # claim draw is set to be true to allign with normal tournament rules
            is_repetition = self.board.is_repetition(3)
            is_50_move_rule = self.board.can_claim_fifty_moves()
            is_claimable_draw = is_repetition or is_50_move_rule
            game_over = is_claimable_draw or is_stale_or_checkmate

            if game_over:
                result = self.board.result(claim_draw=True)
                result_val = chess_utils.result_to_int(result)
                self.set_game_result(result_val)
            else:
                self.infos[current_agent] = {'legal_moves': []}
                self.infos[next_agent] = {'legal_moves': next_legal_moves}
                assert len(self.infos[next_agent]['legal_moves'])

        if observe:
            next_observation = self.observe(next_agent)
        else:
            next_observation = None
        return next_observation

    def render(self, mode='human'):
        self.has_rendered = True
        print(self.board)

    def close(self):
        if not self.has_rendered:
            EnvLogger.warn_close_unrendered_env()
        self.has_rendered = False



import pygame
from pygame.locals import QUIT, MOUSEBUTTONUP

from .board import Board
from .board_ui import Board as BoardUI


def manual_control(**kwargs):
    from .tictactoe import env as _env
    env = _env(**kwargs)

    pygame.init()
    clock = pygame.time.Clock()

    # BoardUI to draw the actual board
    board_ui = BoardUI()

    done = all(env.dones.values())

    while not done:
        for event in pygame.event.get():
            if event.type == QUIT:
                done = True
            elif event.type == MOUSEBUTTONUP:
                x, y = event.pos
                action, square = board_ui.get_square_at_pixel(x, y)

                # check if its a valid action
                if env.board.squares[action] == 0:
                    if env.agent_selection == env.agents[0]:
                        square.mark_x()
                    else:
                        square.mark_o()

                    env.step(action)
                    done = all(env.dones.values())

                    if done:
                        winner = env.board.check_for_winner()
                        board_ui.display_game_over(winner)
                        break
                else:
                    print("Invalid move")

        pygame.display.update()
        clock.tick(30)

    pygame.time.wait(2000)
    board_ui.close()
